# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause
"""
Domain definitions for LBM
"""
import logging
import sys
import copy
import numpy as np
import mpi4py.MPI as mpi

from .geometry import Geometry
from .stencil import Stencil
from .mpi_topology import MpiTopology
from .validator import validate
from . import viewer
from .utils import hsl_to_rgb

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def _create_view_something(something, unvtot):
    """
    The user can pass in parameters of the visualize function
    a boolean or a list or a tuple of integers

    This function returns
    a list (eventually empty) of the considered velocities

    Examples
    --------

    > _create_view_something(True, 3)
        [0, 1, 2]

    > _create_view_something(False, 3)
        []

    > _create_view_something(1, 3)
        [1]

    > _create_view_something([0, 2, 3], 3)
        [0, 2]

    """
    if isinstance(something, bool):
        if something:
            return list(range(unvtot))
        else:
            return []
    elif isinstance(something, int):
        # pylint: disable=chained-comparison
        if something >= 0 and something < unvtot:
            return [something, ]
        else:
            return []
    elif isinstance(something, (list, tuple)):
        lst_something = []
        for s_k in something:
            # pylint: disable=chained-comparison
            if s_k >= 0 and s_k < unvtot:
                lst_something.append(s_k)
        return lst_something
    else:
        s = "Error in visualize (domain): \n"
        s += "optional parameter view_distance and view_bound should be\n"
        s += "  boolean, integer, list or tuple\n"
        log.error(s)
        return []


def _fix_color(vk):
    """
    fix the color of the plot
    for a given velocity
    """
    return hsl_to_rgb(
        np.sin(vk[0]+17)*np.cos(vk[1]+18)*np.cos(vk[2]+12),
        1, 0.5
    )


class Domain:
    """
    Create a domain that defines the fluid part and the solid part
    and computes the distances between these two states.

    Parameters
    ----------

    dico : dictionary
        that contains the following `key:value`

            - box : a dictionary that defines the computational box
            - elements : the list of the elements
                (available elements are given in the module
                :ref:`elements <mod_elements>`)
            - space_step : the spatial step
            - schemes : a list of dictionaries,

        each of them defining a elementary
        :py:class:`Scheme <pylbm.Scheme>`
        we only need the velocities to define a domain

    need_validation : bool
        boolean to specify if the dictionary has to be validated (optional)

    Notes
    -----

    The dictionary that defines the box should contains
    the following `key:value`

    - x : a list of the bounds in the first direction
    - y : a list of the bounds in the second direction (optional)
    - z : a list of the bounds in the third direction (optional)
    - label : an integer or a list of integers
      (length twice the number of dimensions)
      used to label each edge (optional)

    See :py:class:`Geometry <pylbm.geometry.Geometry>` for more details.

    In 1D, distance[q, i] is the distance between the point x[i]
    and the border in the direction of the qth velocity.

    In 2D, distance[q, j, i] is the distance between the point
    (x[i], y[j]) and the border in the direction of qth
    velocity

    In 3D, distance[q, k, j, i] is the distance between the point
    (x[i], y[j], z[k]) and the border in the direction of qth
    velocity

    In 1D, flag[q, i] is the flag of the border reached by the point
    x[i] in the direction of the qth velocity

    In 2D, flag[q, j, i] is the flag of the border reached by the point
    (x[i], y[j]) in the direction of qth velocity

    In 2D, flag[q, k, j, i] is the flag of the border reached by the point
    (x[i], y[j], z[k]) in the direction of qth velocity

    Warnings
    --------

    the sizes of the box must be a multiple of the space step dx

    Attributes
    ----------

    dim : int
      number of spatial dimensions (example: 1, 2, or 3)
    globalbounds : ndarray
      the bounds of the box in each spatial direction
    bounds : ndarray
      the local bounds of the process in each spatial direction
    dx : double
      space step (example: 0.1, 1.e-3)
    type : string
      type of data (example: 'float64')
    stencil : Stencil
      the stencil of the velocities (object of the class
      :py:class:`Stencil <pylbm.stencil.Stencil>`)
    global_size : list
      number of points in each direction
    extent : list
      number of points to add on each side (max velocities)
    coords : ndarray
      coordinates of the domain
    in_or_out : ndarray
      defines the fluid and the solid part
      (fluid: value=valin, solid: value=valout)
    distance : ndarray
      defines the distances to the borders.
      The distance is scaled by dx and is not equal to valin only for
      the points that reach the border with the specified velocity.
      shape = (number_of_velocities, nx, ny, nz)
    flag : ndarray
      NumPy array that defines the flag of the border reached with the
      specified velocity
    normal : ndarray
      numpy array containing the normal vector at the boundary points
      reached with the specified velocity.
      shape = (number_of_velocities, nx, ny, nz, dim)
    valin : int
        value in the fluid domain
    valout : int
        value in the solid domain

    Examples
    --------

    >>> dico = {'box': {'x': [0, 1], 'label': 0},
    ...         'space_step': 0.1,
    ...         'schemes': [{'velocities': list(range(3))}],
    ...        }
    >>> dom = Domain(dico)
    >>> dom
    +--------------------+
    | Domain information |
    +--------------------+
        - spatial dimension: 1
        - space step: 0.1
        - with halo:
            bounds of the box: [-0.05] x [1.05]
            number of points: [12]
        - without halo:
            bounds of the box: [0.05] x [0.95]
            number of points: [10]
    <BLANKLINE>
        +----------------------+
        | Geometry information |
        +----------------------+
            - spatial dimension: 1
            - bounds of the box: [0. 1.]

    >>> dico = {'box': {'x': [0, 1], 'y': [0, 1], 'label': [0, 0, 1, 1]},
    ...         'space_step': 0.1,
    ...         'schemes': [{'velocities': list(range(9))},
    ...                     {'velocities': list(range(5))}
    ...                    ],
    ...        }
    >>> dom = Domain(dico)
    >>> dom
    +--------------------+
    | Domain information |
    +--------------------+
        - spatial dimension: 2
        - space step: 0.1
        - with halo:
            bounds of the box: [-0.05 -0.05] x [1.05 1.05]
            number of points: [12, 12]
        - without halo:
            bounds of the box: [0.05 0.05] x [0.95 0.95]
            number of points: [10, 10]
    <BLANKLINE>
        +----------------------+
        | Geometry information |
        +----------------------+
            - spatial dimension: 2
            - bounds of the box: [0. 1.] x [0. 1.]

    see demo/examples/domain/


    """

    def __init__(self, dico, need_validation=True):
        self.valin = 999  # value in the fluid domain
        self.valout = -1   # value in the solid domain

        if dico is not None and need_validation:
            # pylint: disable=undefined-variable
            validate(dico, __class__.__name__)

        self.geom = Geometry(dico, need_validation=False)
        self.stencil = Stencil(dico, need_validation=False)
        self.dx = dico['space_step']
        self.dim = self.geom.dim
        self.compute_normal = True

        self.box_label = copy.copy(self.geom.box_label)

        self.mpi_topo = None
        self.construct_mpi_topology(dico)

        self.global_size = []
        self.create_coords()

        # pylint: disable=no-value-for-parameter
        region = self.mpi_topo.get_region(*self.global_size)

        # Modify box_label if the border becomes an interface
        for i in range(self.dim):
            if region[i][0] != 0:
                self.box_label[2*i] = -2
            if region[i][1] != self.global_size[i]:
                self.box_label[2*i + 1] = -2

        # distance to the borders
        total_size = [self.stencil.unvtot] + self.shape_halo
        vect_total_size = [self.stencil.unvtot] + self.shape_halo + [self.dim]
        self.in_or_out = self.valin*np.ones(self.shape_halo)
        self.distance = self.valin*np.ones(total_size)
        self.flag = self.valin*np.ones(total_size, dtype='int')
        if self.compute_normal:
            self.normal = np.zeros(vect_total_size)
        else:
            self.normal = None

        # compute the distance and the flag for the primary box
        self.__add_init(self.box_label)
        for elem in self.geom.list_elem:
            # treat each element of the geometry
            self.__add_elem(elem)

        self.clean()
        log.info(self.__str__())

    @property
    def shape_halo(self):
        """
        shape of the whole domain with the halo points.
        """
        return [c.size for c in self.coords_halo]

    @property
    def shape_in(self):
        """
        shape of the interior domain.
        """
        return [c.size for c in self.coords]

    @property
    def x(self):
        """
        x component of the coordinates in the interior domain.
        """
        return self.coords[0]

    @property
    def y(self):
        """
        y component of the coordinates in the interior domain.
        """
        return self.coords[1]

    @property
    def z(self):
        """
        z component of the coordinates in the interior domain.
        """
        return self.coords[2]

    @property
    def x_halo(self):
        """
        x component of the coordinates of the whole domain
        (halo points included).
        """
        return self.coords_halo[0]

    @property
    def y_halo(self):
        """
        y component of the coordinates of the whole domain
        (halo points included).
        """
        return self.coords_halo[1]

    @property
    def z_halo(self):
        """
        z component of the coordinates of the whole domain
        (halo points included).
        """
        return self.coords_halo[2]

    def __str__(self):
        from .utils import header_string
        from .jinja_env import env
        template = env.get_template('domain.tpl')
        return template.render(
            header=header_string('Domain information'),
            dom=self
        )

    def __repr__(self):
        return self.__str__()

    def construct_mpi_topology(self, dico):
        """
        Create the mpi topology
        """
        period = [True]*self.dim

        if dico is None:
            comm = mpi.COMM_WORLD
        else:
            comm = dico.get('comm', mpi.COMM_WORLD)
        self.mpi_topo = MpiTopology(self.dim, period, comm)

    def create_coords(self):
        """
        Create the coordinates of the interior domain and the whole domain
        with halo points.
        """
        phys_box = self.geom.bounds  # the physical box where the domain lies

        # validation of the space step with the physical box size
        for k in range(self.dim):
            self.global_size.append((phys_box[k][1] - phys_box[k][0])/self.dx)
            if not self.global_size[-1].is_integer():
                dummy = round(self.global_size[-1])
                diff_n = dummy - self.global_size[-1]
                if abs(diff_n) < self.dx:
                    message = "The length of the box "
                    message += "in the direction {0:d} ".format(k)
                    message += "is not exactly a multiple of the space step\n"
                    message += "The error in the length of the domain "
                    message += "in the direction {0:d} ".format(k)
                    message += "is {0:10.3e}".format(diff_n*self.dx)
                    log.info(message)
                    self.global_size[-1] = dummy
                else:
                    message = "The length of the box "
                    message += "in the direction {0:d} ".format(k)
                    message += "must be a multiple of the space step\n"
                    message += "The number of points is "
                    message += "{0:.15f}".format(self.global_size[-1])
                    log.error(message)
                    sys.exit()

        # we now are sure that global_size item are integers
        self.global_size = np.asarray(self.global_size, dtype='int')
        # pylint: disable=no-value-for-parameter
        region = self.mpi_topo.get_region(*self.global_size)
        region_size = [r[1] - r[0] for r in region]

        # spatial mesh
        halo_size = np.asarray(self.stencil.vmax)
        halo_beg = self.dx*(halo_size - 0.5)

        self.coords_halo = [
            np.linspace(
                phys_box[k][0] + self.dx*region[k][0] - halo_beg[k],
                phys_box[k][0] + self.dx*region[k][1] + halo_beg[k],
                region_size[k] + 2*halo_size[k]
            )
            for k in range(self.dim)
        ]
        # modification to avoid the case
        # with no velocity in one direction
        ind_beg = [k for k in halo_size]
        ind_end = [-k if k > 0 else 1 for k in halo_size]
        self.coords = [
            # self.coords_halo[k][halo_size[k]:-halo_size[k]]
            self.coords_halo[k][ind_beg[k]:ind_end[k]]
            for k in range(self.dim)
        ]

    def get_bounds_halo(self):
        """
        Return the coordinates of the bottom right and upper left corner of the
        whole domain with halo points.
        """
        bottom_right = np.asarray(
            [self.coords_halo[k][0] for k in range(self.dim)]
        )
        upper_left = np.asarray(
            [self.coords_halo[k][-1] for k in range(self.dim)]
        )
        return bottom_right, upper_left

    def get_bounds(self):
        """
        Return the coordinates of the bottom right and upper left corner of the
        interior domain.
        """
        bottom_right = np.asarray([self.coords[k][0] for k in range(self.dim)])
        upper_left = np.asarray([self.coords[k][-1] for k in range(self.dim)])
        return bottom_right, upper_left

    # pylint: disable=too-many-locals
    def __add_init(self, label):
        halo_size = np.asarray(self.stencil.vmax)
        phys_domain = [
            slice(h, -h) if h > 0 else slice(None)
            for h in halo_size
        ]
        phys_domain_vect = [
            slice(h, -h) if h > 0 else slice(None)
            for h in halo_size
        ]
        self.in_or_out[:] = self.valout
        in_view = self.in_or_out[tuple(phys_domain)]
        in_view[:] = self.valin

        # first index: the velocity index
        # following indices: the physical points
        # last index (for vectors): the dimension
        phys_domain.insert(0, slice(None))
        phys_domain_vect.insert(0, slice(None))
        phys_domain_vect.insert(len(phys_domain_vect), slice(None))
        dist_view = self.distance[tuple(phys_domain)]
        flag_view = self.flag[tuple(phys_domain)]
        if self.compute_normal:
            norm_view = self.normal[tuple(phys_domain_vect)]

        def new_indices(dvik, iuv, indices, dist_view):
            new_ind = copy.deepcopy(indices)
            ind = np.where(dist_view[tuple(indices)] > dvik)
            i = 1
            for j in range(self.dim):
                if j != iuv:
                    new_ind[j + 1] = ind[i]
                    i += 1
            return new_ind

        s = self.stencil
        uvels = [s.uvx, s.uvy, s.uvz]
        # loop over the dimension
        for iuvel, uvel in enumerate(uvels[:self.dim]):
            # uvel is the list of all the velocities
            # in the x-direction, then y-direction and z-direction
            # loop over the velocities
            for k, vk in np.ndenumerate(uvel):
                indices = [k] + [slice(None)]*self.dim
                if vk < 0 and label[2*iuvel] != -2:
                    for i in range(-vk):
                        indices[iuvel + 1] = i
                        dvik = -(i + .5)/vk
                        nind = new_indices(
                            dvik, iuvel,
                            indices,
                            dist_view
                        )
                        dist_view[tuple(nind)] = dvik
                        flag_view[tuple(nind)] = label[2*iuvel]
                        if self.compute_normal:
                            norm_view[tuple(nind + [slice(0, iuvel)])] = 0
                            norm_view[tuple(nind + [iuvel])] = -1
                elif vk > 0 and label[2*iuvel + 1] != -2:
                    for i in range(vk):
                        indices[iuvel + 1] = -i - 1
                        dvik = (i + .5)/vk
                        nind = new_indices(
                            dvik, iuvel,
                            indices,
                            dist_view
                        )
                        dist_view[tuple(nind)] = dvik
                        flag_view[tuple(nind)] = label[2*iuvel+1]
                        if self.compute_normal:
                            norm_view[tuple(nind + [slice(0, iuvel)])] = 0
                            norm_view[tuple(nind + [iuvel])] = 1

    # pylint: disable=too-many-locals
    def __add_elem(self, elem):
        """
        Add an element

            - if elem.isfluid = False as a solid part. (bw=0)
            - if elem.isfluid = True as a fluid part.  (bw=1)

        FIX: this function works only for a 2D problem.
             Need to be improved and implement for the 3D.
        """
        # compute the box around the element adding vmax safety points
        vmax = self.stencil.vmax
        elem_bl, elem_ur = elem.get_bounds()
        phys_bl, _ = self.get_bounds_halo()
        tmp = np.array((elem_bl - phys_bl)/self.dx, int) - vmax
        nmin = np.maximum(vmax, tmp)
        tmp = np.array((elem_ur - phys_bl)/self.dx, int) + vmax + 1
        nmax = np.minimum(vmax + self.shape_in, tmp)

        # set the grid
        space_slice = [slice(imin, imax) for imin, imax in zip(nmin, nmax)]
        total_slice = [slice(None)] + space_slice
        total_slice_vect = [slice(None)] + space_slice + [slice(None)]
        # local view of the arrays
        ioo_view = self.in_or_out[tuple(space_slice)]
        dist_view = self.distance[tuple(total_slice)]
        flag_view = self.flag[tuple(total_slice)]
        if self.compute_normal:
            norm_view = self.normal[tuple(total_slice_vect)]

        tcoords = (self.coords_halo[d][s] for d, s in enumerate(space_slice))
        grid = np.meshgrid(*tcoords, sparse=True, indexing='ij')

        if not elem.isfluid:  # add a solid part
            ind_solid = elem.point_inside(grid)
            ind_fluid = np.logical_not(ind_solid)
            ioo_view[ind_solid] = self.valout
        else:  # add a fluid part
            ind_fluid = elem.point_inside(grid)
            ind_solid = np.logical_not(ind_fluid)
            ioo_view[ind_fluid] = self.valin

        for k in range(self.stencil.unvtot):
            vk = np.asarray(self.stencil.unique_velocities[k].v)
            if np.any(vk != 0):
                space_slice = [
                    slice(imin + vk[d], imax + vk[d])
                    for imin, imax, d in zip(nmin, nmax, range(self.dim))
                ]
                # check the cells that are out
                # when we move with the vk velocity
                out_cells = self.in_or_out[tuple(space_slice)] == self.valout
                # compute the distance and set the boundary label
                # of each cell and the element with the vk velocity
                alpha, border, normvect = elem.distance(
                    grid, self.dx*vk, 1.,
                    self.compute_normal
                )
                # take the indices where the distance is lower than 1
                # between a fluid cell and the border of the element
                # with the vk velocity
                indx = np.logical_and(alpha > 0, ind_fluid)
                if out_cells.size != 0:
                    indx = np.logical_and(indx, out_cells)

                if elem.isfluid:
                    # take all points in the fluid in the ioo_view
                    indfluidinbox = ioo_view == self.valin
                    # take all the fluid points in the box
                    # (not only in the created element)
                    # which always are in fluid after a displacement
                    # of the velocity vk
                    border_to_interior = np.logical_and(
                        np.logical_not(out_cells), indfluidinbox
                    )
                    dist_view[k][border_to_interior] = self.valin
                    flag_view[k][border_to_interior] = self.valin
                else:
                    dist_view[k][ind_solid] = self.valin
                    flag_view[k][ind_solid] = self.valin

                # set distance
                ind4 = np.where(indx)
                if not elem.isfluid:
                    ind3 = np.where(alpha[ind4] < dist_view[k][ind4])[0]
                else:
                    ind3 = np.where(
                        np.logical_or(
                            alpha[ind4] > dist_view[k][ind4],
                            dist_view[k][ind4] == self.valin
                        )
                    )[0]

                ind = [i[ind3] for i in ind4]
                dist_view[k][tuple(ind)] = alpha[tuple(ind)]
                flag_view[k][tuple(ind)] = border[tuple(ind)]
                if self.compute_normal:
                    for i in range(self.dim):
                        norm_view[k][tuple(ind + [i])] = - normvect[
                            tuple(ind + [i])
                        ]

    def clean(self):
        """
        clean the domain when multiple elements are added
        some unused distances or normal vectors have been computed
        """
        # loop of the unique velocity
        # look for the outer points where the distance is computed
        # fix these points to full outer points
        for k in range(self.stencil.unvtot):
            indk = np.logical_and(
                self.distance[k] > 0,
                self.in_or_out == self.valout
            )
            self.distance[k][indk] = self.valin
            self.flag[k][indk] = self.valin
            if self.compute_normal:
                self.normal[k][indk] = 0

    def list_of_labels(self):
        """
        Get the list of all the labels used in the geometry.
        """
        labels = np.unique(self.box_label)
        return np.union1d(labels, self.geom.list_of_elements_labels())

    # pylint: disable=too-complex
    def visualize(self,
                  viewer_app=viewer.matplotlib_viewer,
                  view_geom=True,
                  view_distance=False,
                  view_in=True,
                  view_out=True,
                  view_bound=False,
                  view_normal=False,
                  label=None,
                  scale=1):
        """
        Visualize the domain by creating a plot.

        Parameters
        ----------
        viewer_app : Viewer, optional
            define the viewer to plot the domain
            default is viewer.matplotlib_viewer
        view_geom : boolean
            view the underlying geometry
            default is True
        view_distance : boolean or int or list, optional
            view the distance between the interior points and the border
            default is False
            if True, then all velocities are considered
            can specify some specific velocities in a list
        view_in : boolean, optional
            view the inner points
            default is True
        view_out : boolean, optional
            view the outer points
            default is True
        view_bound : boolean or int or list, optional
            view the points on the bounds
            default is False
        view_normal : boolean or int or list, optional
            view the normal vectors
            default is False
        label : int or list, optional
            view the distance only for the specified labels
        scale : int or float, optional
            scale used for the symbol (default 1)

        Returns
        -------

        object
            views

        """

        fig = viewer_app.Fig(dim=self.dim)
        view = fig[0]
        view.title = "Domain"

        # fix the axis for all dimensions
        delta_l = .25 * max(
            np.diff(self.geom.bounds, axis=1).flatten()
        )  # small length to fix the empty area around the figure
        bornes = [*[-delta_l, delta_l]]*max(2, self.dim)
        bornes[:2*self.dim] += self.geom.bounds.flatten()
        view.axis(*bornes, dim=self.dim, aspect='equal')

        if view_geom and self.dim == 2:
            color_fluid = (0.9, 0.9, 0.9)
            xmin, xmax = self.geom.bounds[0][:]
            ymin, ymax = self.geom.bounds[1][:]
            view.polygon(
                np.array(
                    [
                        [xmin, ymin],
                        [xmin, ymax],
                        [xmax, ymax],
                        [xmax, ymin]
                    ]
                ),
                color_fluid, alpha=1
            )
            for elem in self.geom.list_elem:
                if elem.isfluid:
                    color = color_fluid
                else:
                    color = (1., 1., 1.)
                elem.visualize(
                    view, color,
                    alpha=1
                )
        if view_geom and self.dim == 3:
            color_fluid = (0.9, 0.9, 0.9)
            for elem in self.geom.list_elem:
                if elem.isfluid:
                    color = color_fluid
                else:
                    color = (.5, .5, .5)
                elem.visualize(
                    view, color,
                    alpha=0.25
                )

        # compute the size of the symbols for the plot
        # in 1d and 2d: 100*dx
        # in 3d: 10*dx*dx
        coeff = 1 + 9 * (self.dim <= 2)
        size = scale * 10 * coeff * self.dx**(1+(self.dim == 3))

        # view_distance : list of the concerned velocities
        view_distance = _create_view_something(
            view_distance, self.stencil.unvtot
        )
        # view_bound : list of the concerned velocities
        view_bound = _create_view_something(
            view_bound, self.stencil.unvtot
        )
        # view_normal : list of the concerned velocities
        view_normal = _create_view_something(
            view_normal, self.stencil.unvtot
        )

        # temporary boolean array for considering specific labels
        if isinstance(label, int):
            label = (label,)

        def get_inorout_points(val):
            # returns the coordinates where the value is equal to val
            ind = np.where(self.in_or_out == val)
            data = np.zeros((ind[0].size, 3))
            for i in range(self.dim):
                data[:, i] = self.coords_halo[i][ind[i]]
            return data

        # visualize the inner points
        if view_in:
            view.markers(
                get_inorout_points(self.valin),
                size, symbol='o', alpha=.25, color='navy',
                dim=self.dim
            )

        # visualize the outer points
        if view_out:
            view.markers(
                get_inorout_points(self.valout),
                size, symbol='s', alpha=.5, color='orange',
                dim=self.dim
            )

        def get_bounds(label, k, compute_normal=False):
            # returns
            # 1. the indices of the bounds
            #    for the kth velocity and for the given labels
            # 2. the corresponding distances
            # 3. the corresponding normal vectors
            #    if compute_normal is True
            if label is not None:
                dummy = np.zeros(self.distance.shape[1:])
                for labelk in label:
                    dummy += self.flag[k, :] == labelk
                dummy *= self.distance[k, :] <= 1
                indbord = np.where(dummy)
            else:
                indbord = np.where(self.distance[k, :] <= 1)
            if indbord:
                data = np.zeros((indbord[0].size, max(2, self.dim)))
                for i in range(self.dim):
                    data[:, i] = self.coords_halo[i][indbord[i]]
                dist = self.distance[k][indbord]
                if compute_normal:
                    normal = self.normal[k][indbord]
                    return data, dist, normal
                else:
                    return data, dist
            else:
                if compute_normal:
                    return None, 0, None
                else:
                    return None, 0

        # visualize the distance as small lines
        for k in view_distance:
            bound, dist = get_bounds(label, k)
            if bound.size != 0:
                vk = self.stencil.unique_velocities[k].v_full
                color = _fix_color(vk)
                # pylint: disable=unsubscriptable-object
                lines = np.empty((2*bound.shape[0], max(2, self.dim)))
                lines[::2, :] = bound
                lines[1::2, :] = bound \
                    + self.dx*np.outer(dist, vk[:max(2, self.dim)])
                view.segments(lines, alpha=0.5, width=2, color=color)

        # visualize the bounds as diamond
        for k in view_bound:
            bound, dist = get_bounds(label, k)
            if bound.size != 0:
                vk = self.stencil.unique_velocities[k].v_full
                color = np.array([[*_fix_color(vk)]])
                lines = bound + self.dx*np.outer(dist, vk[:max(2, self.dim)])
                view.markers(lines, size, symbol='d', color=color)

        # visualize the normal vectors as small lines
        for k in view_normal:
            bound, dist, normal = get_bounds(label, k, compute_normal=True)
            if bound.size != 0:
                vk = self.stencil.unique_velocities[k].v_full
                color = _fix_color(vk)
                # pylint: disable=unsubscriptable-object
                lines = np.empty((2*bound.shape[0], max(2, self.dim)))
                lines[::2, :] = bound \
                    + self.dx*np.outer(dist, vk[:max(2, self.dim)])
                lines[1::2, :self.dim] = lines[::2, :self.dim] \
                    + self.dx*normal
                view.segments(lines, alpha=0.5, width=1, color=color)

        fig.show()
        return fig



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Useful functions to compute the distance
between a point and an object
"""
# pylint: disable=invalid-name

import logging
import numpy as np


log = logging.getLogger(__name__)  # pylint: disable=invalid-name
tgv = 1.e16  # very large value


def intersection_two_lines(p1, v1, p2, v2):
    """
    intersection between two lines defined by a point and a vector.
    """
    alpha = beta = None
    det = v1[1]*v2[0] - v1[0]*v2[1]
    if det != 0:
        invdet = 1./det
        c1 = p2[0] - p1[0]
        c2 = p2[1] - p1[1]
        alpha = (-v2[1]*c1 + v2[0]*c2)*invdet
        beta = (-v1[1]*c1 + v1[0]*c2)*invdet
    else:
        message = "Problem in intersection_two_lines\n"
        message += "The lines are parallel"
        log.info(message)
    return alpha, beta


def distance_lines(x, y, v, p, vt, dmax, label, normal):
    """
    return distance for several lines

    normal is a boolean
    if normal is True, the normal vector is also returned
    """
    # Warning: the shape can be
    #   - (x.size, y.size) for the 2D geometries
    #   - x.shape = y.shape for the 3D geometries
    if x.shape[1] == 1 and y.shape[0] == 1:
        shape = (x.size, y.size)
    else:
        shape = x.shape
    v2 = np.asarray([x, y], dtype=object)
    alpha = tgv*np.ones(shape)
    border = -np.ones(shape)
    normal_x = np.zeros(shape)
    normal_y = np.zeros(shape)
    for i, vti in enumerate(vt):
        tmp1, tmp2 = intersection_two_lines(v2, v, p[i], vti)
        if tmp1 is not None:
            if dmax is None:
                ind = np.logical_and(
                    tmp1 > 0,
                    np.logical_and(
                        tmp2 >= 0,
                        tmp2 <= 1
                    )
                )
            else:
                ind = np.logical_and(
                    np.logical_and(
                        tmp1 > 0,
                        tmp1 <= dmax
                    ),
                    np.logical_and(
                        tmp2 >= 0,
                        tmp2 <= 1
                    )
                )
            ind = np.where(np.logical_and(alpha > tmp1, ind))
            alpha[ind] = tmp1[ind]
            border[ind] = label[i]
            if normal:
                # compute the normal vector
                # with a negative scalar product with v
                if v[0]*vti[1] < v[1]*vti[0]:
                    sign = 1
                else:
                    sign = -1
                nv = np.sqrt(vti[0]**2+vti[1]**2)
                normal_x[ind] = sign*vti[1]/nv
                normal_y[ind] = -sign*vti[0]/nv
    alpha[alpha == tgv] = -1.
    if normal:
        normal_vect = np.zeros(tuple(list(shape) + [2]))
        normal_vect[..., 0] = normal_x
        normal_vect[..., 1] = normal_y
    else:
        normal_vect = None
    return alpha, border, normal_vect


# pylint: disable=too-many-locals
def distance_ellipse(x, y, v, center, v1, v2, dmax, label, normal):
    """
    return the distance according
    a line defined by a point x, y and a vector v
    to an ellipse defined by a point c and two vectors v1 and v2

    normal is a boolean
    if normal is True, the normal vector is also returned
    """
    # build the equation of the ellipse
    # then write the second order equation in d
    # a d**2 + b d + c = 0
    # delta = b**2-4ac
    X = x - center[0]
    Y = y - center[1]
    # equation of the ellipse
    # vy2 XX + vx2 YY - vxy XY = d
    vx2 = v1[0]**2 + v2[0]**2
    vy2 = v1[1]**2 + v2[1]**2
    vxy = 2 * (v1[0]*v1[1] + v2[0]*v2[1])
    d = (v1[0]*v2[1]-v1[1]*v2[0])**2
    a = v[0]**2*vy2 + v[1]**2*vx2 - v[0]*v[1]*vxy
    b = 2*X*v[0]*vy2 + 2*Y*v[1]*vx2 - (X*v[1]+Y*v[0])*vxy
    c = X**2*vy2 + Y**2*vx2 - X*Y*vxy - d
    delta = b**2 - 4*a*c
    ind = delta >= 0
    delta[ind] = np.sqrt(delta[ind])

    # Warning: the shape can be
    #   - (x.size, y.size) for the 2D geometries
    #   - x.shape = y.shape for the 3D geometries
    shape = ind.shape  # shape of the outputs
    d1 = tgv*np.ones(shape)
    d2 = tgv*np.ones(shape)
    if a != 0:
        d1[ind] = (-b[ind] - delta[ind])/(2*a)
        d2[ind] = (-b[ind] + delta[ind])/(2*a)
    d1[d1 < 0] = tgv
    d2[d2 < 0] = tgv
    d = -np.ones(shape)
    d[ind] = np.minimum(d1[ind], d2[ind])
    d[d == tgv] = -1

    alpha = -np.ones(shape)
    border = -np.ones(shape)
    if dmax is None:
        ind = d > 0
    else:
        ind = np.logical_and(d > 0, d <= dmax)
    alpha[ind] = d[ind]
    border[ind] = label[0]
    if normal:  # compute the normal vector
        # initialization
        normal_vect = np.zeros(tuple(list(shape) + [2]))
        normal_x = normal_vect[..., 0]
        normal_y = normal_vect[..., 1]
        norm_normal = np.ones(shape)
        # coordinates of the point on the ellipse
        Xe = X + alpha * v[0]
        Ye = Y + alpha * v[1]
        # compute the direction
        normal_x[ind] = 2 * Xe[ind] * vy2 - Ye[ind] * vxy
        normal_y[ind] = 2 * Ye[ind] * vx2 - Xe[ind] * vxy
        # compute the sense (opposite to v)
        sense = np.asarray(
            normal_x*v[0] + normal_y*v[1] > 0
        ).nonzero()
        normal_x[sense] *= -1
        normal_y[sense] *= -1
        # normalization
        norm_normal[ind] = np.sqrt(normal_x[ind]**2 + normal_y[ind]**2)
        # copy in the output vector
        normal_x /= norm_normal
        normal_y /= norm_normal
    else:
        normal_vect = None
    return alpha, border, normal_vect


# pylint: disable=too-many-locals
def distance_ellipsoid(x, y, z, v, center, v1, v2, v3, dmax, label, normal):
    """
    return the distance according
    a line defined by a point x, y, z and a vector v
    to an ellipsoid defined by a point c and three vectors v1, v2, and v3

    normal is a boolean
    if normal is True, the normal vector is also returned
    """
    shape = (x.size, y.size, z.size)  # shape of the outputs
    # build the equation of the ellipsoid
    # then write the second order equation in d
    # a d**2 + b d + c = 0
    # delta = b**2-4ac
    X = x - center[0]
    Y = y - center[1]
    Z = z - center[2]
    v12 = np.cross(v1, v2)
    v23 = np.cross(v2, v3)
    v31 = np.cross(v3, v1)
    d = np.inner(v1, v23)**2
    # equation of the ellipsoid
    # cxx XX + cyy YY + czz ZZ + cxy XY + cyz YZ + czx ZX = d
    cxx = v12[0]**2 + v23[0]**2 + v31[0]**2
    cyy = v12[1]**2 + v23[1]**2 + v31[1]**2
    czz = v12[2]**2 + v23[2]**2 + v31[2]**2
    cxy = 2 * (v12[0]*v12[1] + v23[0]*v23[1] + v31[0]*v31[1])
    cyz = 2 * (v12[1]*v12[2] + v23[1]*v23[2] + v31[1]*v31[2])
    czx = 2 * (v12[2]*v12[0] + v23[2]*v23[0] + v31[2]*v31[0])
    a = cxx*v[0]**2 + cyy*v[1]**2 + czz*v[2]**2 \
        + cxy*v[0]*v[1] + cyz*v[1]*v[2] + czx*v[2]*v[0]
    b = (2*cxx*v[0]+cxy*v[1]+czx*v[2])*X \
        + (2*cyy*v[1]+cyz*v[2]+cxy*v[0])*Y \
        + (2*czz*v[2]+czx*v[0]+cyz*v[1])*Z
    c = cxx*X**2 + cyy*Y**2 + czz*Z**2 \
        + cxy*X*Y + cyz*Y*Z + czx*Z*X - d
    delta = b**2 - 4*a*c
    ind = delta >= 0  # wird but it works
    delta[ind] = np.sqrt(delta[ind])

    d1 = 1e16*np.ones(shape)
    d2 = 1e16*np.ones(shape)
    d1[ind] = (-b[ind] - delta[ind])/(2*a)
    d2[ind] = (-b[ind] + delta[ind])/(2*a)
    d1[d1 < 0] = 1e16
    d2[d2 < 0] = 1e16
    d = -np.ones(shape)
    d[ind] = np.minimum(d1[ind], d2[ind])
    d[d == 1e16] = -1

    alpha = -np.ones(shape)
    border = -np.ones(shape)
    if dmax is None:
        ind = d > 0
    else:
        ind = np.logical_and(d > 0, d <= dmax)
    alpha[ind] = d[ind]
    border[ind] = label[0]

    if normal:  # compute the normal vector
        # initialization
        normal_vect = np.zeros(tuple(list(shape) + [3]))
        normal_x = normal_vect[:, :, :, 0]
        normal_y = normal_vect[:, :, :, 1]
        normal_z = normal_vect[:, :, :, 2]
        norm_normal = np.ones(shape)
        # coordinates of the point on the ellipsoid
        Xe = X + alpha * v[0]
        Ye = Y + alpha * v[1]
        Ze = Z + alpha * v[2]
        # compute the direction
        normal_x[ind] = 2*cxx*Xe[ind] + cxy*Ye[ind] + czx*Ze[ind]
        normal_y[ind] = 2*cyy*Ye[ind] + cxy*Xe[ind] + cyz*Ze[ind]
        normal_z[ind] = 2*czz*Ze[ind] + czx*Xe[ind] + cyz*Ye[ind]
        # compute the sense (opposite to v)
        sense = np.asarray(
            normal_x*v[0] + normal_y*v[1] + normal_z*v[2] > 0
        ).nonzero()
        normal_x[sense] *= -1
        normal_y[sense] *= -1
        normal_z[sense] *= -1
        # normalization
        norm_normal[ind] = np.sqrt(
            normal_x[ind]**2 + normal_y[ind]**2 + normal_z[ind]**2
        )
        # copy in the output vector
        normal_x /= norm_normal
        normal_y /= norm_normal
        normal_z /= norm_normal
    else:
        normal_vect = None
    return alpha, border, normal_vect



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Symbolic computation of equivalent equations
"""

#pylint: disable=invalid-name
import sympy as sp
from ..symbolic import alltogether

class EquivalentEquation:
    def __init__(self, scheme):
        # TODO: add source terms

        t, x, y, z = sp.symbols("t x y z", type='real')
        consm = list(scheme.consm.keys())
        nconsm = len(scheme.consm)

        self.consm = sp.Matrix(consm)
        self.dim = scheme.dim

        space = [x, y, z]

        LA = scheme.la

        func = []
        for i in range(nconsm):
            func.append(sp.Function('f{}'.format(i))(t, x, y, z)) #pylint: disable=not-callable
        func = sp.Matrix(func)

        sublist = [(i, j) for i, j in zip(consm, func)]
        sublist_inv = [(j, i) for i, j in zip(consm, func)]

        eq_func = sp.Matrix(scheme.EQ[nconsm:]).subs(sublist)
        s = sp.Matrix(scheme.s[nconsm:])

        all_vel = scheme.stencil.get_all_velocities()

        Lambda = []
        for i in range(all_vel.shape[1]):
            # FIXME: hack for a bug in sympy
            l = [int(v) for v in all_vel[:, i]]
            vd = LA*sp.diag(*l)
            # vd = LA*sp.diag(*all_vel[:, i])
            Lambda.append(scheme.M*vd*scheme.invM)

        phi1 = sp.zeros(s.shape[0], 1) #pylint: disable=unsubscriptable-object
        inv_s = [1/v for v in s]
        sigma = sp.diag(*inv_s) - sp.eye(len(s))/2
        gamma_1 = sp.zeros(nconsm, 1)

        self.coeff_order1 = []
        for dim, lambda_ in enumerate(Lambda):
            A, B = sp.Matrix([lambda_[:nconsm, :nconsm]]), sp.Matrix([lambda_[:nconsm, nconsm:]])
            C, D = sp.Matrix([lambda_[nconsm:, :nconsm]]), sp.Matrix([lambda_[nconsm:, nconsm:]])

            self.coeff_order1.append(A*func + B*eq_func)
            alltogether(self.coeff_order1[-1], nsimplify=True)
            for i in range(nconsm):
                gamma_1[i] += sp.Derivative(self.coeff_order1[-1][i], space[dim])

            dummy = -C*func - D*eq_func
            alltogether(dummy, nsimplify=True)
            for i in range(dummy.shape[0]):
                phi1[i] += sp.Derivative(dummy[i], space[dim])

        self.coeff_order2 = [[sp.zeros(nconsm) for _ in range(scheme.dim)] for _ in range(scheme.dim)]

        for dim, lambda_ in enumerate(Lambda):
            A, B = sp.Matrix([lambda_[:nconsm, :nconsm]]), sp.Matrix([lambda_[:nconsm, nconsm:]])

            meq = sp.Matrix(scheme.EQ[nconsm:])
            jac = meq.jacobian(consm)
            jac = jac.subs(sublist)
            delta1 = jac*gamma_1

            phi1_ = phi1 + delta1
            sphi1 = B*sigma*phi1_
            sphi1 = sphi1.doit()
            alltogether(sphi1, nsimplify=True)

            for i in range(scheme.dim):
                for jc in range(nconsm):
                    for ic in range(nconsm):
                        self.coeff_order2[dim][i][ic, jc] += sphi1[ic].expand().coeff(sp.Derivative(func[jc], space[i]))

        for ic, c in enumerate(self.coeff_order1):
            self.coeff_order1[ic] = c.subs(sublist_inv).doit()

        for ic, c in enumerate(self.coeff_order2):
            for jc, cc in enumerate(c):
                self.coeff_order2[ic][jc] = cc.subs(sublist_inv).doit()

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('equivalent_equation.tpl')
        t, x, y, z, U, Fx, Fy, Fz, Delta = sp.symbols('t, x, y, z, U, Fx, Fy, Fz, Delta_t')
        Bxx, Bxy, Bxz = sp.symbols('Bxx, Bxy, Bxz')
        Byx, Byy, Byz = sp.symbols('Byx, Byy, Byz')
        Bzx, Bzy, Bzz = sp.symbols('Bzx, Bzy, Bzz')

        phys_equation = sp.Derivative(U, t) + sp.Derivative(Fx, x)
        if self.dim > 1:
            phys_equation += sp.Derivative(Fy, y)
        if self.dim == 3:
            phys_equation += sp.Derivative(Fz, z)

        order2 = []
        space = [x, y, z]
        B = [[Bxx, Bxy, Bxz],
             [Byx, Byy, Byz],
             [Bzx, Bzy, Bzz],
            ]

        phys_equation_rhs = 0
        for i in range(self.dim):
            for j in range(self.dim):
                order2.append(sp.pretty(sp.Eq(B[i][j], -Delta*self.coeff_order2[i][j], evaluate=False)))
                phys_equation_rhs += sp.Derivative(B[i][j]*sp.Derivative(U, space[j]), space[i])
        return template.render(header=header_string('Equivalent Equations'),
                               dim=self.dim,
                               phys_equation=sp.pretty(sp.Eq(phys_equation, phys_equation_rhs)),
                               conserved_moments=sp.pretty(sp.Eq(U, self.consm, evaluate=False)),
                               order1=[sp.pretty(sp.Eq(F, coeff, evaluate=False)) for F, coeff in zip([Fx, Fy, Fz][:self.dim], self.coeff_order1)],
                               order2=order2
                              )

    def __repr__(self):
        return self.__str__()

    def vue(self):
        import jinja2

        try:
            import ipyvuetify as v
            import ipywidgets as widgets
        except ImportError:
            raise ImportError("Please install ipyvuetify")

        t, x, y, z, U, Fx, Fy, Fz, Delta = sp.symbols('t, x, y, z, U, F_x, F_y, F_z, Delta_t')
        Bxx, Bxy, Bxz = sp.symbols('B_{xx}, B_{xy}, B_{xz}')
        Byx, Byy, Byz = sp.symbols('B_{yx}, B_{yy}, B_{yz}')
        Bzx, Bzy, Bzz = sp.symbols('B_{zx}, B_{zy}, B_{zz}')

        phys_equation = sp.Derivative(U, t) + sp.Derivative(Fx, x)
        if self.dim > 1:
            phys_equation += sp.Derivative(Fy, y)
        if self.dim == 3:
            phys_equation += sp.Derivative(Fz, z)

        order2 = []
        space = [x, y, z]
        B = [[Bxx, Bxy, Bxz],
             [Byx, Byy, Byz],
             [Bzx, Bzy, Bzz],
            ]

        phys_equation_rhs = 0
        for i in range(self.dim):
            for j in range(self.dim):
                phys_equation_rhs += sp.Derivative(B[i][j]*sp.Derivative(U, space[j]), space[i])

        order1_dict = {}
        F = [Fx, Fy, Fz]
        for d in range(self.dim):
            order1_dict[sp.latex(F[d])] = [sp.latex(c) for c in self.coeff_order1[d]]

        order0_template = jinja2.Template("""
        {%- macro coeff(order) %}
            {%- for o in order %}
                $$ {{ o }} $$
            {% endfor %}
        {%- endmacro %}
        {{ coeff(consm) }}
        """)

        order1_template = jinja2.Template("""
        {%- macro coeff_dict(consm, order) %}
            \\begin{align*}
            {%- for key, value in order.items() %}
                {%- for i in range(consm|length) %}
                    {{ key }}^{ {{ consm[i] }} } &= {{ value[i] }} \\\\ \\\\
                {% endfor %}
            {% endfor %}
            \\end{align*}
        {%- endmacro %}
        {{ coeff_dict(consm, order1_dict) }}
        """)

        order2_template = jinja2.Template("""
        {%- macro coeff_dict_2(consm, order) %}
            \\begin{align*}
            {%- for key, value in order.items() %}
                {%- for i in range(consm|length) %}
                    {%- for j in range(consm|length) %}
                        {{ key }}^{ {{ consm[i] }}, {{ consm[j] }} } &= {{ value[i*(consm|length) + j] }} \\\\ \\\\
                    {% endfor %}
                {% endfor %}
            {% endfor %}
            \\end{align*}
        {%- endmacro %}
        {{ coeff_dict_2(consm, order2_dict) }}
        """)

        order2_dict = {}
        for i in range(self.dim):
            for j in range(self.dim):
                order2_dict[sp.latex(B[i][j])] = [sp.latex(-Delta*c) for c in self.coeff_order2[i][j]]

        consm = [sp.latex(c) for c in self.consm]
        return v.Container(children=[
            v.Row(children=['The equivalent equation is given by']),
            v.Row(children=[
                widgets.HTMLMath(sp.latex(sp.Eq(phys_equation, phys_equation_rhs), mode='equation*'))
                ],
                justify='center',
            ),
            v.ExpansionPanels(children=[
                v.ExpansionPanel(children=[
                    v.ExpansionPanelHeader(children=['Conserved moments'], class_="title"),
                    v.ExpansionPanelContent(children=[
                        v.Row(children=[
                            widgets.HTMLMath(order0_template.render(consm=consm))
                            ],
                            justify='center'
                        )
                    ])
                ], class_="ma-2"),
                v.ExpansionPanel(children=[
                    v.ExpansionPanelHeader(children=['Order 1'], class_="title"),
                    v.ExpansionPanelContent(children=[
                        v.Row(children=[
                            widgets.HTMLMath(order1_template.render(consm=consm, order1_dict=order1_dict))
                            ],
                            justify='center'
                        )
                    ])
                ], class_="ma-2"),
                v.ExpansionPanel(children=[
                    v.ExpansionPanelHeader(children=['Order 2'], class_="title"),
                    v.ExpansionPanelContent(children=[
                        v.Row(children=[
                            widgets.HTMLMath(order2_template.render(consm=consm, order2_dict=order2_dict))
                            ],
                            justify='center'
                        )
                    ])
                ], class_="ma-2"),
            ])
        ])

    def _repr_mimebundle_(self, **kwargs):
        data = {
            'text/plain': repr(self),
        }
        data['application/vnd.jupyter.widget-view+json'] = {
            'version_major': 2,
            'version_minor': 0,
            'model_id': self.vue()._model_id
        }
        return data



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Symbolic module
"""

import sys
import inspect
import numpy as np
import sympy as sp
from sympy.matrices.common import ShapeError

# List of symbols used in pylbm
nx, ny, nz, nv = sp.symbols("nx, ny, nz, nv", integer=True) #pylint: disable=invalid-name
ix, iy, iz, iv = sp.symbols("ix, iy, iz, iv", integer=True) #pylint: disable=invalid-name
ix_, iy_, iz_, iv_ = sp.symbols("ix_, iy_, iz_, iv_", integer=True) #pylint: disable=invalid-name
rel_ux, rel_uy, rel_uz = sp.symbols('rel_ux, rel_uy, rel_uz', real=True) #pylint: disable=invalid-name

class SymbolicVector(sp.Matrix):

    def __add__(self, other):
        if np.isscalar(other):
            mat = [a + other for a in self]
            return SymbolicVector._new(self.rows, self.cols, mat, copy=False)
        elif isinstance(other, (np.ndarray, SymbolicVector)):
            if self.shape[0] != other.shape[0]:
                raise ShapeError("SymbolicVector size mismatch: %s + %s." % (self.shape, other.shape))
            if len(other.shape) > 1:
                for s in other.shape[1:]:
                    if s != 1:
                        raise ShapeError("SymbolicVector size mismatch: %s + %s." % (self.shape, other.shape))
            mat = [a + b for a,b in zip(self, other)]
            return SymbolicVector._new(self.rows, self.cols, mat, copy=False)
        else:
            return super(SymbolicVector, self).__add__(other)

    def __radd__(self, other):
        return self + other

    def _multiply(self, other, method):
        if np.isscalar(other):
            mat = [a*other for a in self]
            return SymbolicVector._new(self.rows, self.cols, mat, copy=False)
        elif isinstance(other, (np.ndarray, SymbolicVector, sp.MatrixSymbol)):
            if self.shape[0] != other.shape[0]:
                raise ShapeError("SymbolicVector size mismatch: %s * %s." % (self.shape, other.shape))
            if len(other.shape) > 1:
                for s in other.shape[1:]:
                    if s != 1:
                        raise ShapeError("SymbolicVector size mismatch: %s * %s." % (self.shape, other.shape))
            mat = [a*b for a,b in zip(self, other)]
            return SymbolicVector._new(self.rows, self.cols, mat, copy=False)
        else:
            return getattr(super(SymbolicVector, self), method)(other)

    def __mul__(self, other):
        return self._multiply(other, '__mul__')

    def __rmul__(self, other):
        return self._multiply(other, '__rmul__')

def set_order(array, priority=None):
    """
    Reorder an array according to priority (lower to greater).

    Parameters
    ----------

    array : list
        array to reorder

    priority : list
        define how to reorder (lower to greater)
        (default is None)

    Return
    ------

    list
        new reordered list
    """
    if priority:
        out = []
        for p in np.argsort(priority):
            out.append(array[p])
        return out
    else:
        return array


def indexed(name, shape, index=[iv, ix, iy, iz], velocities=None,
            velocities_index=None, priority=None):
    """
    Return a SymPy matrix or an expression of indexed
    objects.

    Parameters
    ----------

    name : str
        name of the SymPy IndexedBase

    shape : list
        shape of the SymPy IndexedBase

    index : list
        indices of the indexed object
        (default is [iv, ix, iy, iz])

    velocities : list
        apply this list of integers on the space indices
        (default is None)

    velocities_index : list
        list of velocities used in the indexed output
        (default is None)

    priority : list
        define how to reorder the indeices (lower to greater)
        (default is None)

    Return
    ------

    sympy.Matrix or sympy.IndexedBase

    Examples
    --------

    >>> import sympy as sp
    >>> i, j, k = sp.symbols('i, j, k')
    >>> m = indexed("m", [10, 100, 200], [i, j, k])
    >>> m
    m[i, j, k]
    >>> m.shape
    (10, 100, 200)

    >>> m = indexed("m", [10, 100, 200], [i, j, k], velocities_index=range(4))
    >>> m
    Matrix([
    [m[0, j, k]],
    [m[1, j, k]],
    [m[2, j, k]],
    [m[3, j, k]]])

    >>> m = indexed("m", [10, 100, 200], [i, j, k], velocities_index=range(4), priority=[1, 2, 0])
    >>> m
    Matrix([
    [m[k, 0, j]],
    [m[k, 1, j]],
    [m[k, 2, j]],
    [m[k, 3, j]]])

    >>> m = indexed("m", [10, 100, 200], [i, j, k], velocities=[[0, 0], [1, 0], [-1, -1]])
    >>> m
    Matrix([
    [        m[0, j, k]],
    [    m[1, j + 1, k]],
    [m[2, j - 1, k - 1]]])

    """
    if velocities_index and velocities:
        raise ValueError("velocities and velocities_index can't be defined together.")

    output = sp.IndexedBase(name, set_order(shape, priority))

    if velocities_index:
        ind = [set_order([k] + list(index[1:]), priority) for k in velocities_index]
        return SymbolicVector([output[i] for i in ind])
    elif velocities is not None:
        ind = []
        indices = index[1:]
        for iv, v in enumerate(velocities): #pylint: disable=invalid-name
            tmp_ind = []
            for ik, k in enumerate(v): #pylint: disable=invalid-name
                tmp_ind.append(indices[ik] + int(k))
            ind.append(set_order([iv] + tmp_ind, priority))
        return SymbolicVector([output[i] for i in ind])
    else:
        return output[set_order(index, priority)]


def space_idx(ranges, priority=None):
    """
    Return a list of SymPy Idx with the bounds of the ranges.
    This list can be permuted if priority is defined.

    Parameters
    ----------

    ranges : list
        bounds of the range for each created SymPy Idx

    priority : list
        define how to reorder the ranges (lower to greater)
            (default is None)

    Return
    ------

    list
        list of SymPy Idx with the right ranges ordered by priority

    Examples
    --------

    >>> loop = space_idx([(0, 10), (-10, 10)])
    >>> loop
    [ix_, iy_]
    >>> loop[0].lower, loop[0].upper
    (0, 10)
    >>> loop = space_idx([(0, 10), (-10, 10)], priority=[1, 0])
    >>> loop
    [iy_, ix_]
    >>> loop[0].lower, loop[0].upper
    (-10, 10)

    """
    indices = [ix_, iy_, iz_]

    idx = []
    for ir, r in enumerate(ranges): #pylint: disable=invalid-name
        idx.append(sp.Idx(indices[ir], r))

    if priority:
        return set_order(idx, priority)
    else:
        return idx

def alltogether(M, nsimplify=False):
    """
    Simplify all the elements of sympy matrix M

    Parameters
    ----------

    M : sympy matrix
       matrix to simplify

    """
    for i in range(M.shape[0]):
        for j in range(M.shape[1]):
            if nsimplify:
                M[i, j] = M[i, j].expand().together().factor().nsimplify()
            else:
                M[i, j] = M[i, j].expand().together().factor()

def recursive_sub(expr, replace):
    for _ in range(len(replace)):
        new_expr = expr.subs(replace)
        if new_expr == expr:
            return new_expr
        else:
            expr = new_expr
    return new_expr

def getargspec_permissive(func):
    """
    find in https://github.com/neithere/argh/blob/master/argh/compat.py

    An `inspect.getargspec` with a relaxed sanity check to support Cython.
    Motivation:
        A Cython-compiled function is *not* an instance of Python's
        types.FunctionType.  That is the sanity check the standard Py2
        library uses in `inspect.getargspec()`.  So, an exception is raised
        when calling `argh.dispatch_command(cythonCompiledFunc)`.  However,
        the CyFunctions do have perfectly usable `.func_code` and
        `.func_defaults` which is all `inspect.getargspec` needs.
        This function just copies `inspect.getargspec()` from the standard
        library but relaxes the test to a more duck-typing one of having
        both `.func_code` and `.func_defaults` attributes.
    """
    if inspect.ismethod(func):
        func = func.im_func

    # Py2 Stdlib uses isfunction(func) which is too strict for Cython-compiled
    # functions though such have perfectly usable func_code, func_defaults.
    if not (hasattr(func, "func_code") and hasattr(func, "func_defaults")):
        raise TypeError('{!r} missing func_code or func_defaults'.format(func))

    args, varargs, varkw = inspect.getargs(func.func_code)
    return inspect.ArgSpec(args, varargs, varkw, func.func_defaults)

PY3 = sys.version_info >= (3,)

if PY3:
    from inspect import getfullargspec as getargspec
else:
    getargspec = getargspec_permissive #pylint: disable=invalid-name

def call_genfunction(function, args):
    from .monitoring import monitor
    from .context import queue
    try:
        func_args = function.arg_dict.keys()
        d = {k:args[k] for k in func_args} #pylint: disable=invalid-name
        d['queue'] = queue
    except: #pylint: disable=bare-except
        func_args = getargspec(function).args
        d = {k:args[k] for k in func_args} #pylint: disable=invalid-name
    monitor(function)(**d)



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Base module that describes what we need to define a new viewer.
"""
from abc import ABC, abstractmethod

class Viewer(ABC):
    """
    Base class which describes the methods needed to
    define a new viewer
    """
    @abstractmethod
    def text(self, t, pos, fontsize, color, **kwargs):
        """
        add text on the figure
        """
        pass

    @abstractmethod
    def line(self, pos, width, color, **kwargs):
        """
        add line on th figure
        """
        pass

    @abstractmethod
    def draw(self):
        """
        show figure
        """
        pass

    @abstractmethod
    def clear(self):
        """
        Clear the figure
        """
        pass

    @abstractmethod
    def axis(self, xmin, xmax, ymin, ymax, zmin, zmax):
        """
        define the axis
        """
        pass

    @abstractmethod
    def is3d(self):
        """
        3d support of the viewer
        """
        pass



from sympy.printing.codeprinter import CodePrinter

from sympy.core import sympify, Eq
from sympy.core.basic import Basic
from sympy.core.symbol import Symbol
from ..ast import Assignment

from sympy.printing.precedence import precedence
from sympy.sets.fancysets import Range

# dictionary mapping sympy function to (argument_conditions, C_function).
# Used in CCodePrinter._print_Function(self)
known_functions = {
    "Abs": [(lambda x: not x.is_integer, "fabs")],
    "gamma": "tgamma",
    "sin": "sin",
    "cos": "cos",
    "tan": "tan",
    "asin": "asin",
    "acos": "acos",
    "atan": "atan",
    "atan2": "atan2",
    "exp": "exp",
    "log": "log",
    "erf": "erf",
    "sinh": "sinh",
    "cosh": "cosh",
    "tanh": "tanh",
    "asinh": "asinh",
    "acosh": "acosh",
    "atanh": "atanh",
    "floor": "floor",
    "ceiling": "ceil",
}

reserved_words = ['auto',
                  'if',
                  'break',
                  'int',
                  'case',
                  'long',
                  'char',
                  'register',
                  'continue',
                  'return',
                  'default',
                  'short',
                  'do',
                  'sizeof',
                  'double',
                  'static',
                  'else',
                  'struct',
                  'entry',
                  'switch',
                  'extern',
                  'typedef',
                  'float',
                  'union',
                  'for',
                  'unsigned',
                  'goto',
                  'while',
                  'enum',
                  'void',
                  'const',
                  'signed',
                  'volatile']

class CythonCodePrinter(CodePrinter):
    """A printer to convert python expressions to strings of cython code"""
    printmethod = "_cythoncode"
    language = "Cython"

    _default_settings = {
        'order': None,
        'full_prec': 'auto',
        'precision': 16,
        'user_functions': {},
        'human': True,
        'contract': True,
        'inline': True,
        'dereference': set(),
        'error_on_reserved': False,
        'reserved_word_suffix': '_',
    }

    def __init__(self, settings=None):
        settings = settings or {}
        super(CythonCodePrinter, self).__init__(settings)
        self.known_functions = dict(known_functions)
        userfuncs = settings.get('user_functions', {})
        self.known_functions.update(userfuncs)
        self._dereference = set(settings.get('dereference', []))
        self.reserved_words = set(reserved_words)

    def doprint(self, expr, assign_to=None):
        """
        Print the expression as code.

        Parameters
        ----------
        expr : Expression
            The expression to be printed.

        assign_to : Symbol, MatrixSymbol, or string (optional)
            If provided, the printed code will set the expression to a
            variable with name ``assign_to``.
        """
        from sympy.matrices.expressions.matexpr import MatrixSymbol

        if isinstance(assign_to, str):
            if expr.is_Matrix:
                assign_to = MatrixSymbol(assign_to, *expr.shape)
            else:
                assign_to = Symbol(assign_to)
        elif not isinstance(assign_to, (Basic, type(None))):
            raise TypeError("{0} cannot assign to object of type {1}".format(
                    type(self).__name__, type(assign_to)))

        if assign_to:
            expr = Assignment(assign_to, expr)
        else:
            # _sympify is not enough b/c it errors on iterables
            expr = sympify(expr)

        # keep a set of expressions that are not strictly translatable to Code
        # and number constants that must be declared and initialized
        self._not_supported = set()
        self._number_symbols = set()

        if isinstance(expr, Eq):
            expr = Assignment(expr.lhs, expr.rhs)

        lines = self._print(expr).splitlines()

        # format the output
        if self._settings["human"]:
            frontlines = []
            if len(self._not_supported) > 0:
                frontlines.append(self._get_comment(
                        "Not supported in {0}:".format(self.language)))
                for expr in sorted(self._not_supported, key=str):
                    frontlines.append(self._get_comment(type(expr).__name__))
            for name, value in sorted(self._number_symbols, key=str):
                frontlines.append(self._declare_number_const(name, value))
            lines = frontlines + lines
            lines = self._format_code(lines)
            result = "\n".join(lines)
        else:
            lines = self._format_code(lines)
            result = (self._number_symbols, self._not_supported,
                    "\n".join(lines))
        return result

    def _get_expression_indices(self, expr, assign_to):
        from sympy.tensor.indexed import Idx

        # need to remove not Idx indices !!!
        rinds = expr.atoms(Idx)
        linds = assign_to.atoms(Idx)

        # support broadcast of scalar
        if linds and not rinds:
            rinds = linds
        return self._sort_optimized(rinds, assign_to)

    def _rate_index_position(self, p):
        return p*5

    def _get_statement(self, codestring):
        return "%s" % codestring

    def _get_comment(self, text):
        return "# {0}".format(text)

    def _declare_number_const(self, name, value):
        return "cdef double const {0} = {1}".format(name, value)

    def _format_code(self, lines):
        return self.indent_code(lines)

    def _traverse_matrix_indices(self, mat):
        rows, cols = mat.shape
        return ((i, j) for i in range(rows) for j in range(cols))

    def _print_Pow(self, expr):
        if "Pow" in self.known_functions:
            return self._print_Function(expr)
        PREC = precedence(expr)
        if expr.exp == -1:
            return '1.0/%s' % (self.parenthesize(expr.base, PREC))
        elif expr.exp == 0.5:
            return 'sqrt(%s)' % self._print(expr.base)
        elif expr.exp > 0 and expr.exp.is_integer:
            line = "%s"%(self.parenthesize(expr.base, PREC))
            for i in range(1, expr.exp):
                line += '*' + "%s"%(self.parenthesize(expr.base, PREC))
            return '(%s)' % line
        else:
            return 'pow(%s, %s)' % (self._print(expr.base),
                                 self._print(expr.exp))

    def _print_Rational(self, expr):
        return self._print(expr.evalf(self._settings["precision"]))

    def _print_Indexed(self, expr):
        elem = []
        for i in range(expr.rank):
            elem.append(self._print(expr.indices[i]))
        return "%s[%s]" % (self._print(expr.base.label), ', '.join(elem))

    def _print_Idx(self, expr):
        return self._print(expr.label)

    def _print_Exp1(self, expr):
        return "M_E"

    def _print_Pi(self, expr):
        return 'M_PI'

    def _print_Infinity(self, expr):
        return 'HUGE_VAL'

    def _print_NegativeInfinity(self, expr):
        return '-HUGE_VAL'

    # TODO Loic: fix
    # def _print_If(self, expr):
    #     lines = []
    #     for c, e in expr.statement:
    #         #temp1, temp2, cond = self.doprint(c)
    #         c = c.replace(Eq, AssignmentIf)
    #         lines.append("if %s:"%self._print(c))
    #         for ee in e:
    #             temp1, temp2, output = self.doprint(ee)
    #             lines.append(output)
    #     lines.append("#end")
    #     return "\n".join(lines)

    def _print_Piecewise(self, expr):
        if expr.args[-1].cond != True:
            # We need the last conditional to be a True, otherwise the resulting
            # function may not return a result.
            raise ValueError("All Piecewise expressions must contain an "
                             "(expr, True) statement to be used as a default "
                             "condition. Without one, the generated "
                             "expression may not evaluate to anything under "
                             "some condition.")
        lines = []
        if expr.has(Assignment):
            for i, (e, c) in enumerate(expr.args):
                if i == 0:
                    lines.append("if (%s):" % self._print(c))
                elif i == len(expr.args) - 1 and c == True:
                    lines.append("else:")
                else:
                    lines.append("else if (%s):" % self._print(c))
                code0 = self._print(e)
                lines.append(code0)
                lines.append("}")
            return "\n".join(lines)
        else:
            # The piecewise was used in an expression, need to do inline
            # operators. This has the downside that inline operators will
            # not work for statements that span multiple lines (Matrix or
            # Indexed expressions).
            ecpairs = ["((%s) ? (\n%s\n)\n" % (self._print(c), self._print(e))
                    for e, c in expr.args[:-1]]
            last_line = ": (\n%s\n)" % self._print(expr.args[-1].expr)
            return ": ".join(ecpairs) + last_line + " ".join([")"*len(ecpairs)])

    def _print_ITE(self, expr):
        from sympy.functions import Piecewise
        _piecewise = Piecewise((expr.args[1], expr.args[0]), (expr.args[2], True))
        return self._print(_piecewise)

    def _print_MatrixElement(self, expr):
        return "{0}[{1}]".format(expr.parent, expr.j +
                expr.i*expr.parent.shape[1])

    def _print_Symbol(self, expr):
        name = super(CythonCodePrinter, self)._print_Symbol(expr)
        if name == 'lambda':
            name = 'lambda_'

        if expr in self._dereference:
            return '(*{0})'.format(name)
        else:
            return name

    def _print_For(self, expr):
        lines = []
        index = expr.target
        for i in index:
            lines.append("for %s in range(%s, %s):"%(i.label, i.lower, i.upper))
        for e in expr.body:
            temp1, temp2, addlines = self.doprint(e)
            if isinstance(addlines, str):
                lines.append(addlines)
            else:
                lines += addlines
        for i in index:
            lines.append("#end")
        return "\n".join(lines)

    def _print_Assignment(self, expr):
        from sympy.functions.elementary.piecewise import Piecewise
        from sympy.matrices.expressions.matexpr import MatrixSymbol
        from sympy.matrices import MatrixBase, MatrixSlice
        lhs = expr.lhs
        rhs = expr.rhs
        # We special case assignments that take multiple lines
        if isinstance(expr.rhs, Piecewise):
            # Here we modify Piecewise so each expression is now
            # an Assignment, and then continue on the print.
            expressions = []
            conditions = []
            for (e, c) in rhs.args:
                expressions.append(Assignment(lhs, e))
                conditions.append(c)
            temp = Piecewise(*zip(expressions, conditions))
            return self._print(temp)
        #elif isinstance(lhs, MatrixSymbol):
        elif isinstance(lhs, (MatrixBase, MatrixSymbol, MatrixSlice)):
            # Here we form an Assignment for each element in the array,
            # printing each one.
            lines = []
            for (i, j) in self._traverse_matrix_indices(lhs):
                temp = Assignment(lhs[i, j], rhs[i, j])
                code0 = self._print(temp)
                lines.append(code0)
            return "\n".join(lines)
        else:
            lhs_code = self._print(lhs)
            rhs_code = self._print(rhs)
            # hack to avoid the printing of m[i] = m[i]
            if lhs_code == rhs_code:
                return ""
            return self._get_statement("%s = %s" % (lhs_code, rhs_code))

    # TODO Loic: fix
    # def _print_AssignmentIf(self, expr):
    #     from sympy.functions.elementary.piecewise import Piecewise
    #     from sympy.matrices.expressions.matexpr import MatrixSymbol
    #     from sympy.matrices import MatrixBase
    #     lhs = expr.lhs
    #     rhs = expr.rhs
    #     # We special case assignments that take multiple lines
    #     if isinstance(expr.rhs, Piecewise):
    #         # Here we modify Piecewise so each expression is now
    #         # an Assignment, and then continue on the print.
    #         expressions = []
    #         conditions = []
    #         for (e, c) in rhs.args:
    #             expressions.append(Assignment(lhs, e))
    #             conditions.append(c)
    #         temp = Piecewise(*zip(expressions, conditions))
    #         return self._print(temp)
    #     elif isinstance(lhs, (MatrixBase, MatrixSymbol)):
    #         # Here we form an Assignment for each element in the array,
    #         # printing each one.
    #         lines = []
    #         for (i, j) in self._traverse_matrix_indices(lhs):
    #             temp = Assignment(lhs[i, j], rhs[i, j])
    #             code0 = self._print(temp)
    #             lines.append(code0)
    #         return "\n".join(lines)
    #     else:
    #         lhs_code = self._print(lhs)
    #         rhs_code = self._print(rhs)
    #         # hack to avoid the printing of m[i] = m[i]
    #         if lhs_code == rhs_code:
    #             return ""
    #         return self._get_statement("%s == %s" % (lhs_code, rhs_code))

    def _print_sign(self, func):
        return '((({0}) > 0) - (({0}) < 0))'.format(self._print(func.args[0]))

    def indent_code(self, code):
        """Accepts a string of code or a list of code lines"""

        if isinstance(code, str):
            code_lines = self.indent_code(code.splitlines(True))
            return ''.join(code_lines)

        tab = "    "
        inc_token = (':', ":\n")
        dec_token = ('end\n',)

        code = [ line.lstrip(' \t') for line in code ]


        increase = [ int(any(map(line.endswith, inc_token))) for line in code ]
        decrease = [ int(any(map(line.endswith, dec_token)))
                     for line in code ]

        pretty = []
        level = 0
        for n, line in enumerate(code):
            if line == '' or line == '\n':
                pretty.append(line)
                continue
            level -= decrease[n]
            pretty.append("%s%s" % (tab*level, line))
            level += increase[n]

        return pretty

    def _print_MatrixBase(self, A):
        # Handle zero dimensions:
        if A.rows == 0 or A.cols == 0:
            return 'zeros(%s, %s)' % (A.rows, A.cols)
        elif (A.rows, A.cols) == (1, 1):
            return "[%s]" % A[0, 0]
        elif A.rows == 1:
            return "[%s]" % A.table(self, rowstart='', rowend='', colsep=' ')
        elif A.cols == 1:
            # note .table would unnecessarily equispace the rows
            return "[%s]" % ", ".join([self._print(a) for a in A])
        return "[%s]" % A.table(self, rowstart='', rowend='',
                                rowsep=';\n', colsep=' ')

    _print_Matrix = \
        _print_DenseMatrix = \
        _print_MutableDenseMatrix = \
        _print_ImmutableMatrix = \
        _print_ImmutableDenseMatrix = \
        _print_MatrixBase



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Module for LBM boundary conditions
"""

import collections
import logging
import types
import numpy as np
from sympy import symbols, IndexedBase, Idx, Eq

from .storage import Array

log = logging.getLogger(__name__) #pylint: disable=invalid-name

#pylint: disable=too-few-public-methods
class BoundaryVelocity:
    """
    Indices and distances for the label and the velocity ksym
    """
    def __init__(self, domain, label, ksym):
        # We are looking for the points on the outside that have a speed
        # that goes in (index ksym) on a border labeled by label.
        # We go through all the lattice velocities and determine the inner points
        # that have the symmetric lattice velocity (index k) that comes out
        # then we write in a list with the order of the lattice velocities
        # involved in the schemes:
        # - indices of the corresponding external points
        # - associated distances
        self.label = label
        self.v = domain.stencil.unique_velocities[ksym]
        v = self.v.get_symmetric()
        num = domain.stencil.unum2index[v.num]

        ind = np.where(domain.flag[num] == self.label)
        self.indices = np.array(ind)
        if self.indices.size != 0:
            self.indices += np.asarray(v.v)[:, np.newaxis]
        self.distance = np.array(domain.distance[(num,) + ind])

class Boundary:
    """
    Construct the boundary problem by defining the list of indices on the border and the methods used on each label.

    Parameters
    ----------
    domain : pylbm.Domain
        the simulation domain
    dico : dictionary
        describes the boundaries
            - key is a label
            - value are again a dictionnary with
                + "method" key that gives the boundary method class used (Bounce_back, Anti_bounce_back, ...)
                + "value_bc" key that gives the value on the boundary

    Attributes
    ----------
    bv_per_label : dictionnary
        for each label key, a list of spatial indices and distance define for each velocity the points
        on the domain that are on the boundary.

    methods : list
        list of boundary methods used in the LBM scheme
        The list contains Boundary_method instance.

    """
    #pylint: disable=too-many-locals
    def __init__(self, domain, generator, dico):
        self.domain = domain

        # build the list of indices for each unique velocity and for each label
        self.bv_per_label = {}
        for label in self.domain.list_of_labels():
            if label in [-1, -2]: # periodic or interface conditions
                continue
            dummy_bv = []
            for k in range(self.domain.stencil.unvtot):
                dummy_bv.append(BoundaryVelocity(self.domain, label, k))
            self.bv_per_label[label] = dummy_bv

        # build the list of boundary informations for each stencil and each label
        dico_bound = dico.get('boundary_conditions', {})
        stencil = self.domain.stencil

        istore = collections.OrderedDict() # important to set the boundary conditions always in the same way !!!
        ilabel = {}
        distance = {}
        value_bc = {}
        time_bc = {}

        #pylint: disable=too-many-nested-blocks
        for label in self.domain.list_of_labels():
            if label in [-1, -2]: # periodic or interface conditions
                continue

            value_bc[label] = dico_bound[label].get('value', None)
            time_bc[label] = dico_bound[label].get('time_bc', False)
            methods = dico_bound[label]['method']
            # for each method get the list of points, the labels and the distances
            # where the distribution function must be updated on the boundary
            for k, v in methods.items():
                for inumk, numk in enumerate(stencil.num[k]):
                    if self.bv_per_label[label][stencil.unum2index[numk]].indices.size != 0:
                        indices = self.bv_per_label[label][stencil.unum2index[numk]].indices
                        distance_tmp = self.bv_per_label[label][stencil.unum2index[numk]].distance
                        velocity = (inumk + stencil.nv_ptr[k])*np.ones(indices.shape[1], dtype=np.int32)[np.newaxis, :]
                        ilabel_tmp = label*np.ones(indices.shape[1], dtype=np.int32)
                        istore_tmp = np.concatenate([velocity, indices])
                        if istore.get(v, None) is None:
                            istore[v] = istore_tmp.copy()
                            ilabel[v] = ilabel_tmp.copy()
                            distance[v] = distance_tmp.copy()
                        else:
                            istore[v] = np.concatenate([istore[v], istore_tmp], axis=1)
                            ilabel[v] = np.concatenate([ilabel[v], ilabel_tmp])
                            distance[v] = np.concatenate([distance[v], distance_tmp])

        # for each method create the instance associated
        self.methods = []
        for k in list(istore.keys()):
            self.methods.append(k(istore[k], ilabel[k], distance[k], stencil,
                                  value_bc, time_bc, domain.distance.shape, generator))


#pylint: disable=protected-access
class BoundaryMethod:
    """
    Set boundary method.

    Parameters
    ----------
    FIXME : add parameters documentation

    Attributes
    ----------
    feq : ndarray
        the equilibrium values of the distribution function on the border
    rhs : ndarray
        the additional terms to fix the boundary values
    distance : ndarray
        distance to the border (needed for Bouzidi type conditions)
    istore : ndarray
        indices of points where we store the boundary condition
    ilabel : ndarray
        label of the boundary
    iload : list
        indices of points needed to compute the boundary condition
    value_bc : dictionnary
       the prescribed values on the border

    """
    def __init__(self, istore, ilabel, distance, stencil, value_bc, time_bc, nspace, generator):
        self.istore = istore
        self.feq = np.zeros((stencil.nv_ptr[-1], istore.shape[1]))
        self.rhs = np.zeros(istore.shape[1])
        self.ilabel = ilabel
        self.distance = distance
        self.stencil = stencil
        self.time_bc = {}
        self.value_bc = {}
        for k in np.unique(self.ilabel):
            self.value_bc[k] = value_bc[k]
            self.time_bc[k] = time_bc[k]
        self.iload = []
        self.nspace = nspace
        self.generator = generator

        # used if time boundary
        self.func = []
        self.args = []
        self.f = []
        self.m = []
        self.indices = []

    def fix_iload(self):
        """
        Transpose iload and istore.

        Must be fix in a future version.
        """
        # Fixme : store in a good way and in the right type istore and iload
        for i in range(len(self.iload)):
            self.iload[i] = np.ascontiguousarray(self.iload[i].T, dtype=np.int32)
        self.istore = np.ascontiguousarray(self.istore.T, dtype=np.int32)

    #pylint: disable=too-many-locals
    def prepare_rhs(self, simulation):
        """
        Compute the distribution function at the equilibrium with the value on the border.

        Parameters
        ----------
        simulation : Simulation
            simulation class

        """

        nv = simulation.container.nv
        sorder = simulation.container.sorder
        nspace = [1]*(len(sorder)-1)
        v = self.stencil.get_all_velocities()

        gpu_support = simulation.container.gpu_support

        for key, value in self.value_bc.items():
            if value is not None:
                indices = np.where(self.ilabel == key)
                # TODO: check the index in sorder to be the most contiguous
                nspace[0] = indices[0].size
                k = self.istore[0, indices]

                s = 1 - self.distance[indices]
                coords = tuple()
                for i in range(simulation.domain.dim):
                    x = simulation.domain.coords_halo[i][self.istore[i + 1, indices]]
                    x += s*v[k, i]*simulation.domain.dx
                    x = x.ravel()
                    for j in range(1, simulation.domain.dim): #pylint: disable=unused-variable
                        x = x[:, np.newaxis]
                    coords += (x,)

                m = Array(nv, nspace, 0, sorder, gpu_support=gpu_support)
                m.set_conserved_moments(simulation.scheme.consm)

                f = Array(nv, nspace, 0, sorder, gpu_support=gpu_support)
                f.set_conserved_moments(simulation.scheme.consm)

                args = coords
                if isinstance(value, types.FunctionType):
                    func = value
                elif isinstance(value, tuple):
                    func = value[0]
                    args += value[1]

                if self.time_bc[key]:
                    func(f, m, 0, *args)
                else:
                    func(f, m, *args)

                simulation.equilibrium(m)
                simulation.m2f(m, f)

                if self.generator.backend.upper() == "LOOPY":
                    f.array_cpu[...] = f.array.get()

                self.feq[:, indices[0]] = f.swaparray.reshape((nv, indices[0].size))

                if self.time_bc[key]:
                    self.func.append(func)
                    self.args.append(args)
                    self.f.append(f)
                    self.m.append(m)
                    self.indices.append(indices[0])

    def update_feq(self, simulation):
        t = simulation.t
        nv = simulation.container.nv

        for i in range(len(self.func)):
            self.func[i](self.f[i], self.m[i], t, *self.args[i])
            simulation.equilibrium(self.m[i])
            simulation.m2f(self.m[i], self.f[i])

            if self.generator.backend.upper() == "LOOPY":
                self.f[i].array_cpu[...] = self.f[i].array.get()

            self.feq[:, self.indices[i]] = self.f[i].swaparray.reshape((nv, self.indices[i].size))

    def _get_istore_iload_symb(self, dim):
        ncond = symbols('ncond', integer=True)

        istore = symbols('istore', integer=True)
        istore = IndexedBase(istore, [ncond, dim+1])

        iload = []
        for i in range(len(self.iload)):
            iloads = symbols('iload%d'%i, integer=True)
            iload.append(IndexedBase(iloads, [ncond, dim+1]))
        return istore, iload, ncond

    @staticmethod
    def _get_rhs_dist_symb(ncond):
        rhs = IndexedBase('rhs', [ncond])
        dist = IndexedBase('dist', [ncond])
        return rhs, dist

    def update(self, ff, **kwargs):
        """
        Update distribution functions with this boundary condition.

        Parameters
        ----------

        ff : array
            The distribution functions
        """
        from .symbolic import call_genfunction

        args = self._get_args(ff)
        args.update(kwargs)
        call_genfunction(self.function, args) #pylint: disable=no-member

    #pylint: disable=possibly-unused-variable
    def _get_args(self, ff):
        dim = len(ff.nspace)
        nx = ff.nspace[0]
        if dim > 1:
            ny = ff.nspace[1]
        if dim > 2:
            nz = ff.nspace[2]

        f = ff.array

        for i in range(len(self.iload)):
            exec('iload{i} = self.iload[{i}]'.format(i=i)) #pylint: disable=exec-used

        istore = self.istore
        rhs = self.rhs
        if hasattr(self, 's'):
            dist = self.s
        ncond = istore.shape[0]
        return locals()

    def move2gpu(self):
        """
        Move arrays needed to compute the boundary on the GPU memory.
        """
        if self.generator.backend.upper() == "LOOPY":
            try:
                import pyopencl as cl
                import pyopencl.array #pylint: disable=unused-variable
                from .context import queue
            except ImportError:
                raise ImportError("Please install loo.py")

            self.rhs = cl.array.to_device(queue, self.rhs)
            if hasattr(self, 's'):
                self.s = cl.array.to_device(queue, self.s) #pylint: disable=attribute-defined-outside-init
            self.istore = cl.array.to_device(queue, self.istore)
            for i in range(len(self.iload)):
                self.iload[i] = cl.array.to_device(queue, self.iload[i])

class BounceBack(BoundaryMethod):
    """
    Boundary condition of type bounce-back

    Notes
    ------

    .. plot:: codes/bounce_back.py

    """
    def set_iload(self):
        """
        Compute the indices that are needed (symmertic velocities and space indices).
        """
        k = self.istore[0]
        ksym = self.stencil.get_symmetric()[k][np.newaxis, :]
        v = self.stencil.get_all_velocities()
        indices = self.istore[1:] + v[k].T
        self.iload.append(np.concatenate([ksym, indices]))

    def set_rhs(self):
        """
        Compute and set the additional terms to fix the boundary values.
        """
        k = self.istore[:, 0]
        ksym = self.stencil.get_symmetric()[k]
        self.rhs[:] = self.feq[k, np.arange(k.size)] - self.feq[ksym, np.arange(k.size)]

    #pylint: disable=too-many-locals
    def generate(self, sorder):
        """
        Generate the numerical code.

        Parameters
        ----------
        sorder : list
            the order of nv, nx, ny and nz
        """
        from .generator import For
        from .symbolic import nx, ny, nz, indexed, ix

        ns = int(self.stencil.nv_ptr[-1])
        dim = self.stencil.dim

        istore, iload, ncond = self._get_istore_iload_symb(dim)
        rhs, _ = self._get_rhs_dist_symb(ncond)

        idx = Idx(ix, (0, ncond))
        fstore = indexed('f', [ns, nx, ny, nz], index=[istore[idx, k] for k in range(dim+1)], priority=sorder)
        fload = indexed('f', [ns, nx, ny, nz], index=[iload[0][idx, k] for k in range(dim+1)], priority=sorder)

        self.generator.add_routine(('bounce_back', For(idx, Eq(fstore, fload + rhs[idx]))))

    @property
    def function(self):
        """Return the generated function"""
        return self.generator.module.bounce_back

class BouzidiBounceBack(BoundaryMethod):
    """
    Boundary condition of type Bouzidi bounce-back [BFL01]

    Notes
    ------

    .. plot:: codes/Bouzidi.py

    """
    def __init__(self, istore, ilabel, distance, stencil, value_bc, time_bc, nspace, generator):
        super(BouzidiBounceBack, self).__init__(istore, ilabel, distance, stencil, value_bc, time_bc, nspace, generator)
        self.s = np.empty(self.istore.shape[1])

    def set_iload(self):
        """
        Compute the indices that are needed (symmertic velocities and space indices).
        """
        k = self.istore[0]
        ksym = self.stencil.get_symmetric()[k]
        v = self.stencil.get_all_velocities()

        iload1 = np.zeros(self.istore.shape, dtype=np.int)
        iload2 = np.zeros(self.istore.shape, dtype=np.int)

        mask = self.distance < .5
        iload1[0, mask] = ksym[mask]
        iload2[0, mask] = ksym[mask]
        iload1[1:, mask] = self.istore[1:, mask] + v[k[mask]].T
        iload2[1:, mask] = self.istore[1:, mask] + 2*v[k[mask]].T
        self.s[mask] = 2.*self.distance[mask]

        mask = np.logical_not(mask)
        iload1[0, mask] = ksym[mask]
        iload2[0, mask] = k[mask]
        iload1[1:, mask] = self.istore[1:, mask] + v[k[mask]].T
        iload2[1:, mask] = self.istore[1:, mask] + v[k[mask]].T
        self.s[mask] = .5/self.distance[mask]

        self.iload.append(iload1)
        self.iload.append(iload2)

    def _get_args(self, ff):
        dim = len(ff.nspace)
        nx = ff.nspace[0]
        if dim > 1:
            ny = ff.nspace[1]
        if dim > 2:
            nz = ff.nspace[2]

        f = ff.array
        # FIXME: needed to have the same results between numpy and cython
        # That means that there are dependencies between the rhs and the lhs
        # during the loop over the boundary elements
        # check why (to test it use air_conditioning example)
        fcopy = ff.array.copy()

        for i in range(len(self.iload)):
            exec('iload{i} = self.iload[{i}]'.format(i=i)) #pylint: disable=exec-used

        istore = self.istore
        rhs = self.rhs
        if hasattr(self, 's'):
            dist = self.s
        ncond = istore.shape[0]
        return locals()

    def set_rhs(self):
        """
        Compute and set the additional terms to fix the boundary values.
        """
        k = self.istore[:, 0]
        ksym = self.stencil.get_symmetric()[k]
        self.rhs[:] = self.feq[k, np.arange(k.size)] - self.feq[ksym, np.arange(k.size)]

    #pylint: disable=too-many-locals
    def generate(self, sorder):
        """
        Generate the numerical code.

        Parameters
        ----------
        sorder : list
            the order of nv, nx, ny and nz
        """
        from .generator import For
        from .symbolic import nx, ny, nz, indexed, ix

        ns = int(self.stencil.nv_ptr[-1])
        dim = self.stencil.dim

        istore, iload, ncond = self._get_istore_iload_symb(dim)
        rhs, dist = self._get_rhs_dist_symb(ncond)

        idx = Idx(ix, (0, ncond))
        fstore = indexed('f', [ns, nx, ny, nz], index=[istore[idx, k] for k in range(dim+1)], priority=sorder)
        fload0 = indexed('fcopy', [ns, nx, ny, nz], index=[iload[0][idx, k] for k in range(dim+1)], priority=sorder)
        fload1 = indexed('fcopy', [ns, nx, ny, nz], index=[iload[1][idx, k] for k in range(dim+1)], priority=sorder)

        self.generator.add_routine(('Bouzidi_bounce_back', For(idx, Eq(fstore, dist[idx]*fload0 + (1-dist[idx])*fload1 + rhs[idx]))))

    @property
    def function(self):
        """Return the generated function"""
        return self.generator.module.Bouzidi_bounce_back

class AntiBounceBack(BounceBack):
    """
    Boundary condition of type anti bounce-back

    Notes
    ------

    .. plot:: codes/anti_bounce_back.py

    """
    def set_rhs(self):
        """
        Compute and set the additional terms to fix the boundary values.
        """
        k = self.istore[:, 0]
        ksym = self.stencil.get_symmetric()[k]
        self.rhs[:] = self.feq[k, np.arange(k.size)] + self.feq[ksym, np.arange(k.size)]

    #pylint: disable=too-many-locals
    def generate(self, sorder):
        """
        Generate the numerical code.

        Parameters
        ----------
        sorder : list
            the order of nv, nx, ny and nz
        """
        from .generator import For
        from .symbolic import nx, ny, nz, indexed, ix

        ns = int(self.stencil.nv_ptr[-1])
        dim = self.stencil.dim

        istore, iload, ncond = self._get_istore_iload_symb(dim)
        rhs, _ = self._get_rhs_dist_symb(ncond)

        idx = Idx(ix, (0, ncond))
        fstore = indexed('f', [ns, nx, ny, nz], index=[istore[idx, k] for k in range(dim+1)], priority=sorder)
        fload = indexed('f', [ns, nx, ny, nz], index=[iload[0][idx, k] for k in range(dim+1)], priority=sorder)

        self.generator.add_routine(('anti_bounce_back', For(idx, Eq(fstore, -fload + rhs[idx]))))

    @property
    def function(self):
        return self.generator.module.anti_bounce_back

class BouzidiAntiBounceBack(BouzidiBounceBack):
    """
    Boundary condition of type Bouzidi anti bounce-back

    Notes
    ------

    .. plot:: codes/Bouzidi.py

    """
    def set_rhs(self):
        """
        Compute and set the additional terms to fix the boundary values.
        """
        k = self.istore[:, 0]
        ksym = self.stencil.get_symmetric()[k]
        self.rhs[:] = self.feq[k, np.arange(k.size)] + self.feq[ksym, np.arange(k.size)]

    #pylint: disable=too-many-locals
    def generate(self, sorder):
        """
        Generate the numerical code.

        Parameters
        ----------
        sorder : list
            the order of nv, nx, ny and nz
        """
        from .generator import For
        from .symbolic import nx, ny, nz, indexed, ix

        ns = int(self.stencil.nv_ptr[-1])
        dim = self.stencil.dim

        istore, iload, ncond = self._get_istore_iload_symb(dim)
        rhs, dist = self._get_rhs_dist_symb(ncond)

        idx = Idx(ix, (0, ncond))
        fstore = indexed('f', [ns, nx, ny, nz], index=[istore[idx, k] for k in range(dim+1)], priority=sorder)
        fload0 = indexed('f', [ns, nx, ny, nz], index=[iload[0][idx, k] for k in range(dim+1)], priority=sorder)
        fload1 = indexed('f', [ns, nx, ny, nz], index=[iload[1][idx, k] for k in range(dim+1)], priority=sorder)

        self.generator.add_routine(('Bouzidi_anti_bounce_back', For(idx, Eq(fstore, -dist[idx]*fload0 + (1-dist[idx])*fload1 + rhs[idx]))))

    @property
    def function(self):
        return self.generator.module.Bouzidi_anti_bounce_back

class Neumann(BoundaryMethod):
    """
    Boundary condition of type Neumann

    """
    name = 'neumann'
    def set_rhs(self):
        """
        Compute and set the additional terms to fix the boundary values.
        """
        pass

    def set_iload(self):
        """
        Compute the indices that are needed (symmertic velocities and space indices).
        """
        k = self.istore[0]
        v = self.stencil.get_all_velocities()
        indices = self.istore[1:] + v[k].T
        self.iload.append(np.concatenate([k[np.newaxis, :], indices]))

    #pylint: disable=too-many-locals
    def generate(self, sorder):
        """
        Generate the numerical code.

        Parameters
        ----------
        sorder : list
            the order of nv, nx, ny and nz
        """
        from .generator import For
        from .symbolic import nx, ny, nz, indexed, ix

        ns = int(self.stencil.nv_ptr[-1])
        dim = self.stencil.dim

        istore, iload, ncond = self._get_istore_iload_symb(dim)

        idx = Idx(ix, (0, ncond))
        fstore = indexed('f', [ns, nx, ny, nz], index=[istore[idx, k] for k in range(dim+1)], priority=sorder)
        fload = indexed('f', [ns, nx, ny, nz], index=[iload[0][idx, k] for k in range(dim+1)], priority=sorder)

        self.generator.add_routine((self.name, For(idx, Eq(fstore, fload))))

    @property
    def function(self):
        """Return the generated function"""
        return self.generator.module.neumann

class NeumannX(Neumann):
    """
    Boundary condition of type Neumann along the x direction

    """
    name = 'neumannx'
    def set_iload(self):
        """
        Compute the indices that are needed (symmertic velocities and space indices).
        """
        k = self.istore[0]
        v = self.stencil.get_all_velocities()
        indices = self.istore[1:].copy()
        indices[0] += v[k].T[0]
        self.iload.append(np.concatenate([k[np.newaxis, :], indices]))

    @property
    def function(self):
        """Return the generated function"""
        return self.generator.module.neumannx

class NeumannY(Neumann):
    """
    Boundary condition of type Neumann along the y direction

    """
    name = 'neumanny'
    def set_iload(self):
        """
        Compute the indices that are needed (symmertic velocities and space indices).
        """
        k = self.istore[0]
        v = self.stencil.get_all_velocities()
        indices = self.istore[1:].copy()
        indices[1] += v[k].T[1]
        self.iload.append(np.concatenate([k[np.newaxis, :], indices]))

    @property
    def function(self):
        """Return the generated function"""
        return self.generator.module.neumanny

class NeumannZ(Neumann):
    """
    Boundary condition of type Neumann along the z direction

    """
    name = 'neumannz'
    def set_iload(self):
        """
        Compute the indices that are needed (symmertic velocities and space indices).
        """
        k = self.istore[0]
        v = self.stencil.get_all_velocities()
        indices = self.istore[1:].copy()
        indices[1] += v[k].T[2]
        self.iload.append(np.concatenate([k[np.newaxis, :], indices]))

    @property
    def function(self):
        """Return the generated function"""
        return self.generator.module.neumannz



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
__init__ of the module containing the analysis tools
"""

from .equivalent_equation import EquivalentEquation
from .stability import Stability


# FIXME: make pylint happy !
#pylint: disable=all

import sys
import os
import shutil
import tempfile
from subprocess import STDOUT, CalledProcessError, check_output
import importlib

from .codegen import get_code_generator

class CodeWrapError(Exception):
    pass


class CodeWrapper(object):
    """Base Class for code wrappers"""
    _module_name = "wrapped_module"

    @property
    def filename(self):
        return "%s_%s" % (self._module_name, self._module_counter)

    @property
    def full_path(self):
        return "%s/%s_%s" % (self.workdir, self._module_name, self._module_counter)

    @property
    def module_name(self):
        return "%s_%s" % (self._module_name, self._module_counter)

    def __init__(self, generator, filepath=None, flags=[], generate=True, verbose=False):
        """
        generator -- the code generator to use
        """
        self.generator = generator
        self.filepath = filepath
        self.workdir = self.filepath or tempfile.mkdtemp("_sympy_compile")
        self.flags = flags
        self.generate = generate
        self.verbose = verbose
        self._module_counter = 0


    def _generate_code(self, routines):
        self.generator.write(
            routines, self.full_path, True, True, False)

    def wrap_code(self, routines):
        if not os.access(self.workdir, os.F_OK):
            os.mkdir(self.workdir)

        try:
            sys.path.append(self.workdir)
            if self.generate:
                self._prepare_files(routines)
                self._generate_code(routines)
                self._process_files(routines)
            if self.module_name in sys.modules:
                sys.modules.pop(self.module_name)
            importlib.invalidate_caches()
            mod = importlib.import_module(self.module_name)
        finally:
            sys.path.remove(self.workdir)
            self._module_counter += 1
            if not self.filepath:
                try:
                    shutil.rmtree(self.workdir)
                except OSError:
                    # Could be some issues on Windows
                    pass

        return mod

    def _process_files(self, routine):
        command = self.command
        command.extend(self.flags)

        try:
            retoutput = check_output(command, stderr=STDOUT)
        except CalledProcessError as e:
            raise CodeWrapError(
                "Error while executing command: %s. Command output is:\n%s" % (
                    " ".join(command), e.output.decode()))
        if self.verbose:
            print(retoutput)


class CythonCodeWrapper(CodeWrapper):
    @property
    def command(self):
        bld = open(self.full_path + '.pyxbld', "w")
        code = """
def make_ext(modname, pyxfilename):
    from distutils.extension import Extension

    return Extension(name = modname,
                     sources=[pyxfilename],
                     #extra_compile_args = ['-O3', '-fopenmp, '-w'],
                     #extra_link_args= ['-fopenmp'])
                     extra_compile_args = ['-O3', '-w']
                     #extra_compile_args = ['-O3', '-fopenmp', '-w'],
                     #extra_link_args= ['-fopenmp'])
                    )
                    """
        bld.write(code)
        bld.close()
        build_file = os.path.join(self.workdir, 'build.py')
        bld = open(build_file, 'w')
        code = f"""
import pyximport
pyximport.install(build_dir=r'{self.workdir}', inplace=True)
import {self.filename}
        """
        bld.write(code)
        bld.close()

        if self.verbose:
            print(open(self.full_path + '.pyx').read())

        command = [sys.executable, build_file]

        return command

    def _prepare_files(self, routines):
        pass

class PythonCodeWrapper(CodeWrapper):
    @property
    def command(self):
        return []

    def _prepare_files(self, routines):
        pass

    def _process_files(self, routines):
        if self.verbose:
            print(open(self.full_path + '.py').read())

def get_code_wrapper(backend):
    CodeWrapClass = {"NUMPY" : PythonCodeWrapper,
                     "CYTHON": CythonCodeWrapper,
                     "LOOPY": PythonCodeWrapper}.get(backend.upper())
    if CodeWrapClass is None:
        raise ValueError("Language '%s' is not supported." % backend)
    return CodeWrapClass

def autowrap(routines, backend='cython', tempdir=None, generate=True, args=None, flags=[],
    verbose=False):

    code_generator = get_code_generator(backend, "project")
    CodeWrapperClass = get_code_wrapper(backend)
    code_wrapper = CodeWrapperClass(code_generator, tempdir, flags, generate, verbose)

    return code_wrapper.wrap_code(routines)



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Parallelogram element
"""

# pylint: disable=invalid-name

import logging
# from textwrap import dedent
import numpy as np

from .base import Element
from .utils import distance_lines

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Parallelogram(Element):
    """
    Class Parallelogram

    Parameters
    ----------

    point : list
        the coordinates of the first point of the parallelogram
    vecta : list
        the coordinates of the first vector
    vectb : list
        the coordinates of the second vector
    label : list
        four integers (default [0, 0, 0, 0])
    isfluid : boolean
        - True if the parallelogram is added
        - False if the parallelogram is deleted

    Examples
    --------

    the square [0,1]x[0,1]

    >>> point = [0., 0.]
    >>> vecta = [1., 0.]
    >>> vectb = [0., 1.]
    >>> Parallelogram(point, vecta, vectb)
    +---------------+
    | Parallelogram |
    +---------------+
        - dimension: 2
        - start point: [0. 0.]
        - v1: [1. 0.]
        - v2: [0. 1.]
        - label: [0, 0, 0, 0]
        - type: solid

    Attributes
    ----------

    number_of_bounds : int
        4
    dim: int
        2
    point : ndarray
        the coordinates of the first point of the parallelogram
    v1 : ndarray
        the coordinates of the first vector
    v2 : ndarray
        the coordinates of the second vector
    label : list
        the list of the label of the edge
    isfluid : boolean
        True if the parallelogram is added
        and False if the parallelogram is deleted

    """
    def __init__(self, point, vecta, vectb, label=0, isfluid=False):
        self.number_of_bounds = 4  # number of edges
        self.dim = 2
        self.point = np.asarray(point)
        self.v1 = np.asarray(vecta)
        self.v2 = np.asarray(vectb)
        super(Parallelogram, self).__init__(label, isfluid)
        log.info(self.__str__())

    def get_bounds(self):
        """
        return the bounds of the parallelogram.
        """
        box = np.asarray([
            self.point,
            self.point + self.v1,
            self.point + self.v1 + self.v2,
            self.point + self.v2
        ])
        return np.min(box, axis=0), np.max(box, axis=0)

    def point_inside(self, grid):
        """
        return a boolean array which defines
        if a point is inside or outside of the parallelogram.

        Notes
        -----

        the edges of the parallelogram are considered as inside.

        Parameters
        ----------

        grid : ndarray
            coordinates of the points

        Returns
        -------

        ndarray
            Array of boolean (True inside the parallelogram, False otherwise)

        """
        x, y = grid
        # Barycentric coordinates
        v2 = np.asarray([x - self.point[0], y - self.point[1]], dtype=object)
        invdelta = 1./(self.v1[0]*self.v2[1] - self.v1[1]*self.v2[0])
        u = (v2[0]*self.v2[1] - v2[1]*self.v2[0])*invdelta
        v = (v2[1]*self.v1[0] - v2[0]*self.v1[1])*invdelta
        return np.logical_and(
            np.logical_and(
                u >= 0,
                v >= 0
            ),
            np.logical_and(
                u <= 1,
                v <= 1
            )
        )

    def distance(self, grid, v, dmax=None, normal=False):
        """
        Compute the distance in the v direction between the parallelogram
        and the points defined by (x, y).

        .. image:: ../figures/Parallelogram.png
            :width: 100%

        Parameters
        ----------

        grid : ndarray
            coordinates of the points
        v : ndarray
            direction of interest
        dmax : float
            distance max
        normal : bool
            return the normal vector if True (default False)

        Returns
        -------

        ndarray
            array of distances if normal is False and
            the coordinates of the normal vectors
            if normal is True
        """
        x, y = grid
        # points and triangle edges defining the lines for the intersections
        # with the lines defined by (x, y) and v
        p = [[0, 0], [0, 0], self.v2, self.v1]
        vt = [self.v1, self.v2, self.v1, self.v2]

        return distance_lines(
            x - self.point[0],
            y - self.point[1],
            v, p, vt,
            dmax, self.label, normal
        )

    def __str__(self):
        from ..utils import header_string
        from ..jinja_env import env
        template = env.get_template('square.tpl')
        elem_type = 'fluid' if self.isfluid else 'solid'
        return template.render(
            header=header_string(self.__class__.__name__),
            elem=self, type=elem_type
        )

    def visualize(self,
                  viewer, color, viewlabel=False,
                  scale=np.ones(2), alpha=1.
                  ):
        A = [self.point[k] for k in range(2)]
        B = [A[k] + self.v1[k] for k in range(2)]
        C = [B[k] + self.v2[k] for k in range(2)]
        D = [A[k] + self.v2[k] for k in range(2)]
        viewer.polygon([A, B, C, D], color, alpha=alpha)
        if viewlabel:
            viewer.text(str(self.label[0]), [0.5*(A[0]+B[0]), 0.5*(A[1]+B[1])])
            viewer.text(str(self.label[1]), [0.5*(A[0]+D[0]), 0.5*(A[1]+D[1])])
            viewer.text(str(self.label[2]), [0.5*(C[0]+D[0]), 0.5*(C[1]+D[1])])
            viewer.text(str(self.label[3]), [0.5*(B[0]+C[0]), 0.5*(B[1]+C[1])])



# Authors:
#     Loic Gouarin <loic.gouarin@polytechnique.edu>
#     Benjamin Graille <benjamin.graille@math.u-psud.fr>
#
# License: BSD 3 clause

"""
Module which implements a Cartesian MPI topology
"""

import numpy as np
import mpi4py.MPI as mpi

from .options import options

class MpiTopology:
    """
    Interface construction using a MPI topology.

    Parameters
    ----------

    dim : int
      number of spatial dimensions (1, 2, or 3)
    comm : comm
      the default MPI communicator
    period : list
      boolean list that specifies if a direction is periodic or not.
      Its size is dim.

    Attributes
    ----------

    dim : int
      number of spatial dimensions (1, 2, or 3)
    comm : comm
      the communicator of the topology
    split : tuple
      number of processes in each direction
    neighbors : list
      list of the neighbors where we have to send and to receive messages
    sendType : list
      list of subarrays that defines the part of data to be send
    sendTag : list
      list of tags for the send messages
    recvType : list
      list of subarrays that defines the part of data to update during a receive message
    recvTag : list
      list of tags for the receive messages

    Methods
    -------

    set_options :
      defines command line options.
    get_coords :
      return the coords of the process in the MPI topology.
    set_subarray :
      create subarray for the send and receive message
    update :
      update a numpy array according to the subarrays and the topology.

    """
    def __init__(self, dim, period, comm=mpi.COMM_WORLD):
        self.dim = dim
        self.set_options()

        self.comm = comm
        # if npx, npy and npz are all set to the default value (1)
        # then Compute_dims performs the splitting of the domain
        if self.npx == self.npy == self.npz == 1:
            size = comm.Get_size()
            split = mpi.Compute_dims(size, self.dim)
        else:
            split = (self.npx, self.npy, self.npz)

        self.split = np.asarray(split[:self.dim])
        self.cartcomm = comm.Create_cart(self.split, period)

    def get_region_indices_(self, n, axis=0):
        """
        1D region indices owned by each sub domain.

        Parameters
        ----------

        n : int
            number of total discrete points for a given axis
        axis : int
            axis used in the MPI topology

        Returns
        -------

        list
            list of regions owned by each processes for a given axis

        """
        region_indices = [0]
        nproc = self.cartcomm.Get_topo()[0][axis]
        for i in range(nproc):
            region_indices.append(region_indices[-1] + n//nproc + ((n % nproc) > i))
        return region_indices

    def get_region_indices(self, nx, ny=None, nz=None):
        """
        Region indices owned by each sub domain.

        Parameters
        ----------

        nx : int
            number of total discrete points in x direction
        ny : int
            number of total discrete points in y direction
            default is None
        nz : int
            number of total discrete points in z direction
            default is None

        Returns
        -------

        list
            list of regions owned by each processes

        """
        region_indices = [self.get_region_indices_(nx, 0)]
        if ny is not None:
            region_indices.append(self.get_region_indices_(ny, 1))
        if nz is not None:
            region_indices.append(self.get_region_indices_(nz, 2))
        return region_indices

    def get_coords(self):
        """
        return the coords of the process in the MPI topology
        as a numpy array.
        """
        rank = self.cartcomm.Get_rank()
        return np.asarray(self.cartcomm.Get_coords(rank))

    def get_region(self, nx, ny=None, nz=None):
        """
        Region indices owned by the sub domain.

        Parameters
        ----------

        nx : int
            number of total discrete points in x direction
        ny : int
            number of total discrete points in y direction
            default is None
        nz : int
            number of total discrete points in z direction
            default is None

        Returns
        -------

        list
            region owned by the process

        """

        region_indices = self.get_region_indices(nx, ny, nz)

        coords = self.get_coords()
        region = []
        for i in range(coords.size):
            region.append([region_indices[i][coords[i]],
                           region_indices[i][coords[i] + 1]
                          ])
        return region

    def set_options(self):
        """
        defines command line options.
        """
        self.npx = int(options().npx)
        self.npy = int(options().npy)
        self.npz = int(options().npz)

def get_directions(dim):
    """
    Return an array with all the directions around.

    Parameters
    ----------

    dim : int
      number of spatial dimensions (1, 2, or 3)

    Returns
    -------

    ndarray
        all the possible directions with a stencil of 1

    Examples
    --------

    >>> get_directions(1)
    array([[-1],
           [ 0],
           [ 1]])
    >>> get_directions(2)
    array([[-1, -1],
           [-1,  0],
           [-1,  1],
           [ 0, -1],
           [ 0,  0],
           [ 0,  1],
           [ 1, -1],
           [ 1,  0],
           [ 1,  1]], dtype=int32)

    """
    common_direction = np.array([-1, 0, 1])

    if dim == 1:
        directions = common_direction[:, np.newaxis]
    elif dim == 2:
        common_direction = common_direction[np.newaxis, :]

        directions = np.empty((9, 2), dtype=np.int32)
        directions[:, 0] = np.repeat(common_direction, 3, axis=1).flatten()
        directions[:, 1] = np.repeat(common_direction, 3, axis=0).flatten()
    elif dim == 3:
        common_direction = common_direction[np.newaxis, :]

        directions = np.empty((27, 3), dtype=np.int32)
        directions[:, 0] = np.repeat(common_direction, 9, axis=1).flatten()
        directions[:, 1] = np.repeat(np.repeat(common_direction, 3, axis=0), 3).flatten()
        directions[:, 2] = np.repeat(common_direction, 9, axis=0).flatten()

    return directions
